
# ZRTeXtor.pm : library for processing TeX font data
#
# Written by ZR <zrbabbler@infoseek.jp>

######## start package
package ZRTeXtor;
use strict qw( refs vars subs );
require Exporter;
our $VERSION = 1.000_00;  # 2008/05/13 v1.0.0
our @ISA = qw( Exporter );
our @EXPORT = ();
our %EXPORT_TAGS = (
  'gen' => [qw(
    textool_error
    read_whole_file write_whole_file
    arraymap
    test1 test2 test3
  )],
  'pl' => [qw(
    pl_parse pl_parse_list pl_cook_list
    pl_form pl_form_list
    pl_value pl_set_value pl_set_numtype pl_prefer_hex
    pl_rearrange pl_clone pl_sclone
    pl_header pl_header_vpl pl_header_opl
    pl_fontdimen pl_fontdimen_jpl
    pl_fontdimen_opl_rmt pl_fontdimen_jpl_rmt
    pl_get_metric pl_char_part pl_char_part_jpl
  )],
  'x' => [qw(
    x_captured_exec
    x_tftopl x_uptftopl x_pltotf x_vptovf x_opl2ofm
  )],
  'jcode' => [qw(
    jcode_set jcode_chr jcode_ord
  )],
  'ps' => [qw(
    ps_parse0
  )],
  'enc' => [qw(
    enc_register enc_retrieve enc enc_parse enc_form
    ENCF_NORMAL ENCF_ANNOT ENCF_COMPACT ENCF_256
  )],
  'kpse' => [qw(
    kpse kpse_init
  )],
  'ttf' => [qw(
    ttf_generate_opl ttf_generate_jpl_x ttf_generate_jpl
    ttf_get_metric
  )],
  'tfm' => [qw(
    tfm_reduce_jpl_x tfm_reduce_jpl tfm_reduce
    tfm_reduced_metric tfm_map_rmt
  )],
  'vf' => [qw(
    vf_parse vf_form vf_simple_move_code
    vf_for_mapping vf_mapfont vf_set_maofont
  )],
  'jfm' => [qw(
    jfm_use_uptex_tool jfm_parse jfm_form
  )],
);
our @EXPORT_OK = map { @{$_} } (values %EXPORT_TAGS);
$EXPORT_TAGS{all} = [ @EXPORT_OK ];
use Carp;
use Encode qw(encode decode);
use Data::Dump;
*dumpp = *Data::Dump::dump;

######## general ########

our $binmode = 0;           # always read/write in binary mode
our $errmsg = "(no error message)"; # last error message
use constant { HUGE => 1e20, EPS => 1e-7 };
  # TU = TFM factor, DU = DVI factor
use constant { TU => 1 << 20, DU => 1 << 16, B31 => 1 << 31 };
use constant { M32 => 0xFFFFFFFF, M31 => 0x7FFFFFFF };
# for get_temp_name
our $get_temp_base = '__zrtx';
our (@get_temp_id);
get_temp_name_init();

## error($msg, ...)
# Sets the error message and returns nothing (undef in scalar).
# Usually a defined value is returned in success.
sub error
{
  $errmsg = join('', map { (ref $_) ? '[obj]' : $_ } (@_));
  return;
}

## fatal($msg, ...)
# Intended for internal errors....
sub fatal
{
  return error("!!FATAL(", @_, ")");
}

##<*> textool_error()
# Obtains the error message of the last error. This string
# can be used in error handling of your own style.
sub textool_error
{
  return $errmsg;
}

## get_temp_name()
# Returns a temporary file name which is unique in the process.
# Note: Trailing digits in font names can have special meaning
# in some TeX tools.
sub get_temp_name
{
  ++$get_temp_id[0];
  return join('', $get_temp_base, $$, @get_temp_id[1, 0, 2]);
}

## get_temp_anme_init()
# Initializer for get_temp_name.
sub get_temp_name_init
{
  my ($t); $get_temp_id[0] = 0;
  $t = join('', map { ('a' .. 'z')[int(rand() * 26)] } (0 .. 5));
  @get_temp_id[1, 2] = map { substr($t, $_, 3) } (0, 3);
}

##<*> read_whole_file($fnam, $swbin)
# Reads in file $fnam and returns its content as a string.
# If $swbin or the global flag $binmode is true the file is read
# in binary mode.
sub read_whole_file
{
  my ($fnam, $swbin) = @_; my ($hf, $txt); local ($/);
  (defined $fnam) or return;
  open($hf, '<', $fnam)
    or return error("cannot open for read: $fnam");
  if ($binmode || $swbin) { binmode($hf); }
  $txt = readline($hf);
  close($hf);
  return $txt;
}

##<*> write_whole_file($fnam, $txt, $swbin)
# Creates (or crobbers) the file $fnam and write $txt to it.
# If $swbin or the global flag $binmode  is true it writes in
# binary mode.
sub write_whole_file
{
  my ($fnam, $txt, $swbin) = @_; my ($hf);
  open($hf, '>', $fnam)
    or return error("cannot open for write: $fnam");
  if ($binmode || $swbin) { binmode($hf); }
  print $hf ($txt);
  close($hf);
  return 1;
}

## unpack_num($s)
# Decodes a unsigned number in big-endian string.
sub unpack_num
{
  my ($s) = @_;
  return unpack("N", substr("\0\0\0\0$s", length($s)));
}

## unpack_snum($s)
# Decodes a signed number in big-endian string.
my @snum_bound = (0, 0x80, 0x8000, 0x800000, B31);
sub unpack_snum
{
  my ($s) = @_; my ($b, $v);
  $b = $snum_bound[length($s)];
  $v = unpack("N", substr("\0\0\0\0$s", length($s)));
  return ($v >= $b) ? ($v - $b - $b) : $v;
}

## pack_num($v)
# Encodes a unsigned number in big-endian string, with the length
# prefix. In scalar context the returned values are concatenated.
sub pack_num
{
  my ($v) = @_; my ($t, $l);
  ($t = pack('N', $v)) =~ s/^\0{1,3}//; $l = length($t);
  return (wantarray) ? ($l, $t) : (chr($l) . $t);
}

## pack_snum($v)
# Signed version of pack_num.
sub pack_snum
{
  my ($v) = @_; my ($t, $l);
  ($t = pack('N', $v)) =~
    s/^\0{1,3}([\x00-\x7f])|^\xff{1,3}([\x80-\xff])/$+/;
  $l = length($t);
  return (wantarray) ? ($l, $t) : (chr($l) . $t);
}

## signed($v)
# Converts 'unsigned long' to 'signed long'.
sub signed
{
  my ($v) = @_; $v &= M32;
  return ($v >= B31) ? ($v - B31 - B31) : $v;
}

## round($v)
# Rounds a real value to an integer.
sub round
{
  my ($v) = @_;
  return int($v + (($v < 0) ? -0.5 : +0.5));
}

##<*> arraymap($map, $swmm)
# Converts a code map in hash form into array form.
sub arraymap
{
  my ($map, $swmm) = @_; my ($sc, $dc, @u, @amap);
  if (ref $map eq 'HASH') {
    @u = sort { $a <=> $b } (keys %$map);
    foreach $sc (@u) {
      $dc = $map->{$sc};
      if (ref $dc eq 'ARRAY') {
        if ($swmm) {
          foreach (@$dc) { push(@amap, [$sc, $_]); }
        } else { push(@amap, [$sc, $dc->[0]]); }
      } elsif (defined $dc) { push(@amap, [$sc, $dc]); }
    }
    return \@amap;
  } elsif (ref $map eq 'ARRAY') { return $map; }
  else { return; }
}

## squote($str)
# S-quotes a string.
sub squote
{
  my ($str) = @_; $str =~ s/([\\\'])/\\$1/g;
  return "'$str'";
}

## zdquote($str)
# ZD-quotes a string.
sub zdquote
{
  my ($str) = @_; $str =~ s/([\\\"])/\\$1/g;
  $str =~ s/([^\x20-\x7e])/sprintf("\\%02X",$1)/ge;
  return "\"$str\"";
}

######## 'x' section ########

use IPC::Open3; # for open3()

##<*> x_captured_exec(@cmd);
# Spawns the command @cmd, captures its stdout and stderr into
# strings, and returns them.
# --- Am I doing right?
sub x_captured_exec
{
  my (@cmd) = @_; my ($pid, @fs, @ds);
  local(*CHIN, *CHOUT, *CHERR, $/);
  L1:{
    @fs = (get_temp_name(), get_temp_name());
    open(CHOUT, '+>', $fs[0]) or last;
    open(CHERR, '+>', $fs[1]) or last;
    ($pid = open3(\*CHIN, '>&CHOUT', '>&CHERR', @cmd)) or last;
    waitpid($pid, 0);
    seek(CHOUT, 0, 0); $ds[0] = <CHOUT>;
    seek(CHERR, 0, 0); $ds[1] = <CHERR>;
  }
  close(CHIN); close(CHOUT); close(CHERR);
  unlink(@fs);
  return (@ds);
}

##<*> x_tftopl
# Wrapper for 'tftopl' command.
sub x_tftopl
{
  my ($tfm, $cmd) = @_; my ($ftmp, $ftfm, $cout, $cerr);
  if (!defined $cmd) { $cmd = 'tftopl'; }
  if ($tfm =~ m/\.tfm$/i && $tfm !~ /\0/) { $ftfm = $tfm; }
  else {
    $ftfm = $ftmp = get_temp_name() . ".tfm";
    (write_whole_file($ftmp, $tfm, 1)) or return;
  }
  ($cout, $cerr) = x_captured_exec("$cmd $ftfm");
  if (defined $ftmp) { unlink($ftmp); }
  if ($cout eq '' || $cout =~ /CHANGED!\)\s*$/) {
    return error("tftopl failed: $ftfm");
  }
  return pl_parse($cout);
}

##<*> x_uptftopl
# Wrapper for 'uptftopl' command.
# Use of x_tftopl($tfm, "uptftopl") does not seem to work right.
# Perhaps getting uptftopl to 'output to terminal' confuses
# uptftopl as to handling UTF-8 strings....
sub x_uptftopl
{
  my ($tfm, $cmd) = @_; my ($ftmp, $ftfm, $fpl, $cout, $cerr);
  if (!defined $cmd) { $cmd = 'uptftopl'; }
  if ($tfm =~ m/\.tfm$/i && $tfm !~ /\0/) { $ftfm = $tfm; }
  else {
    $ftfm = $ftmp = get_temp_name() . ".tfm";
    (write_whole_file($ftmp, $tfm, 1)) or return;
  }
  $fpl = get_temp_name() . ".pl";
  ($cout, $cerr) = x_captured_exec("$cmd $ftfm $fpl");
  if (defined $ftmp) { unlink($ftmp); }
  $cout = read_whole_file($fpl); unlink($fpl);
  if ($cout eq '' || $cout =~ /CHANGED!\)\s*$/) {
    return error("tftopl failed: $ftfm");
  }
  return pl_parse($cout);
}

##<*> x_pltotf
# Wrapper for 'pltotf' command.
sub x_pltotf
{
  my ($pl, $cmd) = @_;
  my ($ftmp, $fpl, $cout, $ftfm, $tfm, $tn);
  if (!defined $cmd) { $cmd = 'pltotf'; }
  $tn = get_temp_name(); $ftfm = "$tn.tfm";
  if (ref $pl eq 'ARRAY') { $pl = pl_form($pl); }
  if ($pl =~ m/\.pl$/i) { $fpl = $pl; }
  else {
    $fpl = $ftmp =  "$tn.pl";
    (write_whole_file($ftmp, $pl)) or return;
  }
  $cout = `$cmd $fpl $ftfm`;  # no output to stderr
  if (defined $ftmp) { unlink($ftmp); }
  (-f $ftfm) or return error("pltotf failed: $fpl");
  $tfm = read_whole_file($ftfm, 1); unlink($ftfm);
  foreach (split(/\n/, $cout)) {
    if (m/^\s*$/ || m/^I had to round some / ||
        m/Input file is in kanji /) { next; }
    return error("pltotf failed: $fpl");
  }
  return $tfm;
}

##<*> x_vptovf
# Wrapper for 'vptovf' command.
sub x_vptovf
{
  my ($pl, $cmd) = @_;
  my ($ftmp, $fpl, $cout, $ftfm, $tfm, $fvf, $vf, $tn);
  if (!defined $cmd) { $cmd = 'vptovf'; }
  $tn = get_temp_name(); $ftfm = "$tn.tfm"; $fvf = "$tn.vf";
  if (ref $pl eq 'ARRAY') { $pl = pl_form($pl); }
  if ($pl =~ m/\.vpl$/i) { $fpl = $pl; }
  else {
    $fpl = $ftmp =  "$tn.vpl";
    (write_whole_file($ftmp, $pl)) or return;
  }
  $cout = `$cmd $fpl $fvf $ftfm`;  # no output to stderr
  if (defined $ftmp) { unlink($ftmp); }
  (-f $ftfm && -f $fvf) or return error("vptovf failed: $fpl");
  $tfm = read_whole_file($ftfm, 1); unlink($ftfm);
  $vf = read_whole_file($fvf, 1); unlink($fvf);
  foreach (split(/\n/, $cout)) {
    if (m/^\s*$/ || m/^I had to round some / ||
        m/Input file is in kanji / || m/LIG/) { next; }
    return error("vptovf failed: $fpl");
  }
  return ($vf, $tfm);
}

##<*> x_opl2ofm
# Wrapper for 'opl2ofm' command.
sub x_opl2ofm
{
  my ($pl) = @_; my ($ftmp, $fpl, $cout, $ftfm, $tfm, $tn);
  $tn = get_temp_name(); $ftfm = "$tn.ofm";
  if (ref $pl eq 'ARRAY') { $pl = pl_form($pl); }
  if ($pl =~ m/\.opl$/i) { $fpl = $pl; }
  else {
    $fpl = $ftmp =  "$tn.opl";
    (write_whole_file($ftmp, $pl)) or return;
  }
  $cout = `opl2ofm $fpl $ftfm`;  # no output to stderr
  if (defined $ftmp) { unlink($ftmp); }
  (-f $ftfm) or return error("opl2ofm failed: $fpl");
  $tfm = read_whole_file($ftfm, 1); unlink($ftfm);
  foreach (split(/\n/, $cout)) {
    if (m/^\s*$/ || m/^I had to round some / ||
        m/Input file is in kanji /) { next; }
    return error("opl2ofm failed: $fpl");
  }
  return $tfm;
}


######## 'pl' section ########

####---- Handling General Property Lists

# * Data structure: pl-list, pl-struct
# A PL Text consists of a sequence of Lispish lists (without the
# outermost parentheses). In 'raw' structures, Lispish lists are
# represented by Perl arrays with straightforward conversion
# (all atomic elements are strings). Such an array is  called
# 'pl-list' here. The entire PL is represented as an array of
# pl-lists, called 'pl-struct' here.
# However, in 'cooked' structures, some numerical data are
# interpreted. For example, raw data
#    'D', '0100'   (a part of a pl-list)
# are cooked they change to
#    [ CNUM, 100, 'D' ], '0100'
# and moreover the assignment of 500 to this part results in
#    [ CNUM, 500, 'D' ], undef
# And its conversion back into text is 'O 764', because 500 cannot
# be expressed in 'D' notation.

use constant CNUM => ' ';   # flag signifying a cooked nubmer
our $freenum = 0;

my %pl_facecode = (         # 'F' notation
  MRR => 0, MIR => 1, BRR => 2, BIR => 3, LRR => 4, LIR => 5,
  MRC => 6, MIC => 7, BRC => 8, BIC => 9, LRC => 10, LIC => 11,
  MRE => 12, MIE => 13, BRE => 14, BIE => 15, LRE => 16, LIE => 17
);
my %pl_facecode_rev = (reverse %pl_facecode);

##<*> pl_parse($txt)
# Converts a PL text $txt to a 'cooked' pl-struct.
sub pl_parse
{
  my ($txt) = @_; my ($pl, $ent);
  (defined($pl = pl_parse_list("($txt)"))) or return;
  (pl_cook_list($pl)) or return;
  foreach $ent (@$pl) {
    if (!ref $ent) { return error("bareword found: ", $ent); }
  }
  return $pl;
}

##<*> pl_parse_list($txt)
# Converts a text $txt of a Lispish list to a 'raw' pl-list.
sub pl_parse_list
{
  my ($txt) = @_; my (@toks, $pp);
  @toks = grep { $_ ne "" } (split(/(\()|(\))|\s+/, $txt));
  if ($toks[0] ne '(') { return error("missing paren at top"); }
  $pp = pl_corres_paren(\@toks, 0);
  if ($pp == $#toks) {
    return pl_parse_sub(\@toks, 1, $pp - 1);
  } elsif ($pp < 0) {
    return error("unmatched parens (end at level ", -$pp, ")");
  } else {
    return error("unmatched parens (extra tokens)");
  }
}

## pl_parse_sub(...)
# Subcontractor of pl_parse_list.
sub pl_parse_sub
{
  my ($toks, $sp, $ep) = @_; my (@pl, $tok, $p, $pp, $pl2);
  for ($p = $sp; $p <= $ep; $p++) {
    $tok = $toks->[$p];
    if ($tok eq '(') {
      $pp = pl_corres_paren($toks, $p);
      ($p < $pp && $pp <= $ep) or return fatal(0);
      (defined($pl2 = pl_parse_sub($toks, $p + 1, $pp - 1))) or return;
      push(@pl, $pl2); $p = $pp;
    } else {
      push(@pl, $tok);
    }
  }
  return \@pl;
}

## pl_corres_paren($toks, $p)
# Returns the index of the ')' token which corresponds with the
# '(' token at index $p in an array $toks.
sub pl_corres_paren
{
  my ($toks, $p) = @_; my ($tok, $lv);
  for ($lv = 1, ++$p; $p <= $#$toks; $p++) {
    $tok = $toks->[$p];
    if ($tok eq '(') { ++$lv; }
    elsif ($tok eq ')') { --$lv; }
    if ($lv == 0) { last; }
  }
  return ($lv > 0) ? -$lv : $p;
}

##<*> pl_cook_list($pl)
# Makes a raw pl-list $pl cooked.
sub pl_cook_list
{
  my ($pl) = @_; my ($k, $ent, $res);
  for ($k = 0; $k <= $#$pl; $k++) {
    $ent = $pl->[$k];
    if (ref $ent) {
      if ($ent->[0] eq 'COMMENT') {
        splice(@$pl, $k, 1); redo;
      } elsif ($ent->[0] ne CNUM) {
        (pl_cook_list($ent)) or return;
      }
    } elsif ($ent =~ /^[CKDFOHR]$/ && $k < $#$pl) {
      (defined pl_proc_num($pl, $k)) or return;
      ++$k;
    }
  }
  return 1;
}

##<*> pl_form($pl, $ind)
# Converts a pl-struct $pl into a PL text. Here $ind is the amount
# of indent: if negative the result is in in-line form.
sub pl_form
{
  my ($pl, $ind) = @_; my (@cnks, $ent, $txt);
  foreach $ent (@$pl) {
    (defined($txt = pl_form_list($ent, $ind))) or return;
    push(@cnks, $txt);
  }
  if ($ind >= 0) { return join("\n", @cnks, ''); }
  else { return join(' ', @cnks); }
}

# for pl_form_list
my $pl_rx_kent = qr/^[\x80-\xff]|^[JUX][0-9A-Fa-f]{4}$/;

##<*> pl_form_list($pl, $ind)
# Converts a pl-list $pl into a Lispish list.
sub pl_form_list
{
  my ($pl, $ind) = @_; my (@cnks, @lins, @toks);
  my ($k, $t, $lsepp, $lsep, $ent, $tok, $txt);
  if ($ind >= 0) { 
    push(@cnks, '('); $ind += 3;
    $lsepp = $lsep = "\n" . ' ' x $ind;
  } else { push(@cnks, '('); $lsepp = $lsep = ' '; }
  for ($k = 0; $k <= $#$pl; $k++) {
    $ent = $pl->[$k];
    if (ref $ent) {
      if ($ent->[0] eq CNUM) {
        $tok = $pl->[$k + 1]; ++$k;
        if (defined $tok) { push(@lins, $ent->[2], $tok); }
        else {
          (@toks = pl_form_num($ent->[2], $ent->[1])) or return;
          push(@lins, @toks);
        }
      } else {
        if (@lins) { push(@cnks, join(' ', @lins), $lsepp); }
        (defined($txt = pl_form_list($ent, $ind))) or return;
        @lins = (); push(@cnks, $txt, $lsep);
      }
    } elsif ($k > 2 && $ind >= 0 && $ent =~ m/$pl_rx_kent/) {
      if (@lins) { push(@cnks, join(' ', @lins), $lsepp); }
      $t = '';
      while ($pl->[$k] =~ m/$pl_rx_kent/) {
        if (length($t) + length($pl->[$k]) + $ind > 72) { last; }
        $t .= $pl->[$k] . ' '; ++$k;
      }
      --$k; chop($t); @lins = (); push(@cnks, $t, $lsep);
    } else { push(@lins, $ent); }
  }
  push(@cnks, join(' ', @lins), ')');
  if ($ind < 0 && $cnks[-3] eq $lsep) { $cnks[-3] = ''; }
  return join('', @cnks);
}

##<*> pl_value($pl, $k, $sw)
# Reads the number at position $k in pl-list $pl. Note that
# $k is the position of form prefix and not the string represeting
# the nubmer itself.
# The value will be rounded to integers unless $sw is true.
sub pl_value
{
  my ($pl, $k, $sw) = @_; my ($ent);
  $ent = $pl->[$k];
  if (ref $ent && $ent->[0] eq CNUM) {
    return ($sw) ? $ent->[1] : round($ent->[1]);
  }
  return pl_proc_num($pl, $k);
}

## pl_proc_num($pl, $k)
# Converts the number token at position $k in pl-list $pl to
# cooked form.
sub pl_proc_num
{
  my ($pl, $k) = @_; my ($v, $fl, $tok);
  ($fl, $tok) = ($pl->[$k], $pl->[$k + 1]);
  if (defined($v = pl_parse_num($fl, $tok))) {
    $pl->[$k] = [ CNUM, $v, $fl ]; return $v;
  } else {
    return error("malformed number (", $fl, " ", $tok, ")");
  }
}

##<*> pl_set_value($pl, $k, $v)
# Changes the number at position $k in pl-list $pl to $v.
sub pl_set_value
{
  my ($pl, $k, $v) = @_; my ($ent);
  $ent = $pl->[$k];
  (ref $ent && $ent->[0] eq CNUM)
    or return error("illegal assignment", $ent);
  $pl->[$k + 1] = undef; $ent->[1] = $v;
  return 1;
}

##<*> pl_set_real($pl, $k)
# Changes the 'R'-form number at position $k in pl-list $pl to
# $v, which is a non-scaled value.
sub pl_set_real
{
  my ($pl, $k, $v) = @_; my ($ent);
  $ent = $pl->[$k];
  (ref $ent && $ent->[0] eq CNUM && $ent->[2] eq 'R')
    or return error("illegal assignment", $ent);
  $pl->[$k + 1] = undef; $ent->[1] = $v * TU;
  return 1;
}

##<*> pl_set_value($pl, $k, $v)
# Changes the form prefix of the number at position $k in
# pl-list $pl to $v.
sub pl_set_numtype
{
  my ($pl, $k, $v) = @_; my ($ent);
  $ent = $pl->[$k];
  (ref $ent && $ent->[0] eq CNUM)
    or return error("illegal assignment", $ent);
  $pl->[$k + 1] = undef; $ent->[2] = $v;
  return 1;
}

# for pl_prefer_hex()
our $pl_prefer_hex = 0;

##<*> pl_prefer_hex($sw)
# Sets the value of the global flag $pl_prefer_hex. If its value is
# true, then 'H' (instead of 'O') is used as a fallback of 'D'/'C'.
sub pl_prefer_hex
{
  $pl_prefer_hex = $_[0];
}

## pl_parse_num($fl, $tok)
# Converts a number token to the number it expresses. Here $fl is
# a form prefix and $tok is a token,
sub pl_parse_num
{
  my ($fl, $tok) = @_; my ($ll, $ul, $v);
  $ll = 0; $ul = M32;
  if (ref $tok) { return; }
  if ($fl eq 'C') {
    $v = ($tok =~ /^[\x21-\x7e]$/ && ord($tok));
  } elsif ($fl eq 'K') {
    $v = jcode_ord($tok);
  } elsif ($fl eq 'D') {
    $v = ($tok =~ /^-?\d+$/ && $tok); $ul = 255;
  } elsif ($fl eq 'F') {
    $v = (exists $pl_facecode{$tok} && $pl_facecode{$tok});
  } elsif ($fl eq 'O') {
    $v = ($tok =~ /^[0-7]+$/ && oct($tok)); $ul = M32;
  } elsif ($fl eq 'H' || $fl eq 'I') {
    $v = ($tok =~ /^[\da-f]+$/i && hex($tok)); $ul = M32;
  } elsif ($fl eq 'R') {
    $v = ($tok =~ /^[\+\-]?(\d*\.)?\d+$/ && pl_scale_real($tok));
    $ll = - B31; $ul = M31;
  }
  if ($freenum && $fl ne 'R') { $ul = M32; $ll = -$ul; }
  if (!($ll <= $v && $v <= $ul)) { $v = undef; }
  return $v;
}

## pl_form_num($fl, $v)
# Expresses the number $v in form $fl. If the number cannot (or
# should not) be expressed in form $fl, then $fl is fallen back
# to another suitable value. It returns ($fl, $tok) where $tok
# is the resulted expression and $fl is the possibly adapted
# value of form prefix.
sub pl_form_num
{
  my ($fl, $v) = @_; my ($tok);
  $tok = $fl;
  if ($fl eq 'F' && $v > 17) { $fl = 'D'; }
  if ($fl eq 'C' && !pl_isprint($v)) { $fl = 'I'; }
  if ($fl eq 'K' && !pl_isjprint($v)) { $fl = 'H'; }
  if ($fl eq 'D' && $v > 255) { $fl = 'I'; }
  if ($fl eq 'I') { $fl = ($pl_prefer_hex) ? 'H' : 'O'; }
  if (($fl eq 'R' && !(- B31 <= $v && $v <= M31))
      || ($fl ne 'R' && !(0 <= $v && $v <= M32))) {
    #return error("number is out of range ($v for $tok)");
    $v = 1234*TU;
  }
  if ($fl eq 'C') { $tok = chr($v); }
  elsif ($fl eq 'K') { $tok = jcode_chr($v); }
  elsif ($fl eq 'D') { $tok = $v; }
  elsif ($fl eq 'F') { $tok = $pl_facecode_rev{$v}; }
  elsif ($fl eq 'O') { $tok = sprintf("%o", $v); }
  elsif ($fl eq 'H') { $tok = sprintf("%X", $v); }
  elsif ($fl eq 'R') { $tok = pl_form_real($v / TU); }
  return ($fl, $tok);
}

# Note: In PL, big numbers (>255) in decimal are not allowed,
#   whereas they are allowed in VPL.

## pl_form_real($a)
# Expresses a real number in the same way as in PL.
sub pl_form_real
{
  my ($a) = @_; my ($d, @t);
  if ($a < -0.5 / TU) { $a = -$a; @t = ("-"); }
  $a = int($a * TU + 0.5);
  { use integer;
    push(@t, $a / TU, "."); $a %= TU;
    $a = $a * 10 + 5; $d = 10;
    do {
      if ($d > 0x100000) { $a = $a + 0x80000 - $d / 2; }
      push(@t, $a / 0x100000); $a = $a % 0x100000 * 10; $d *= 10;
    } until ($a <= $d);
  }
  return join('', @t);
}

## pl_scale_real($a)
# Returns a real value scaled to tfm-unit, rounded to integer.
sub pl_scale_real
{
  my ($a) = @_;
  return int($a * TU + (($a < 0) ? -0.5 : +0.5));
}


## pl_isprint($c)
# Tests if the number $c is to output really in 'C' form.
sub pl_isprint
{
  my ($c) = @_;
  return (0 <= $c && $c <= 255 && chr($c) =~ /^\w/);
}

## pl_isjprint($c)
# Tests if the number $c is to output in 'K' form.
sub pl_isjprint
{
  my ($c) = @_;
  return defined(jcode_chr($c));
}

####---- Rearranging pl-structs

##<*> pl_rearrange($pl)
# Sorts the pl-lists in a pl-struct $pl so that they are in
# 'usual' order.
sub pl_rearrange
{
  my ($pl) = @_;
  @$pl = sort { pl_prop_pos($a) <=> pl_prop_pos($b) } (@$pl);
  return $pl;
}

# for pl_prop_pos
my %pl_prop_pos_base = (
  DIRECTION         => 0x00010,
  FAMILY            => 0x00011,
  FACE              => 0x00012,
  HEADER            => 0x10000,
  CODINGSCHEME      => 0x20010,
  DESIGNUNITS       => 0x20011,
  DESIGNSIZE        => 0x20012,
  CHECKSUM          => 0x20013,
  SEVENBITSAFEFLAG  => 0x20014,
  FONTDIMEN         => 0x20015,
  BOUNDARYCHAR      => 0x20016,
  MAPFONT           => 0x30000,
  LIGTABLE          => 0x40010,
  GLUEKERN          => 0x40010,
  CHARSINTYPE       => 0x50000,
  CHARACTER         => 0x60000,
  TYPE              => 0x60000,
);

## pl_prop_pos($pl)
# Subcontractor for pl_rearrange.
sub pl_prop_pos
{
  my ($pl) = @_; my ($v);
  $v = $pl_prop_pos_base{$pl->[0]};
  if (!defined $v) { return 0xFFFFF; }
  elsif (($v & 0xffff) == 0) { return $v | pl_value($pl, 1); }
  else { return $v; }
}

####---- Rearranging pl-lists

##<*> pl_clone($pl)
# Returns a deep clone of a pl-list, where the original and the
# clone share no reference.
sub pl_clone
{
  my ($pl) = @_;;
  if (ref $pl eq "ARRAY") {
    return [ map { pl_clone($_) } (@$pl) ];
  } else { return $pl; }
}

##<*> pl_sclone($pl)
# Returns a one-level clone of a pl-list, considering cooked
# number forms which should be uniquified.
sub pl_sclone
{
  my ($pl) = @_;;
  if (ref $pl eq "ARRAY") {
    return [ map {
              (ref $_ eq "ARRAY" && $_->[0] eq CNUM) ? [ @$_ ] : $_
             } (@$pl) ];
  } else { return $pl; }
}


####---- Handling PL/JPL/OPL/VPL Structs

# The three functions below generate a header part (stuffs before
# FONTDIMEN and optionally FONTDIMEN). Here $in is a hash ref with
# the following effective keys:
#   direction, family, vtitle, face, codingscheme, designunits,
#   designsize, checksum, sevenbitsafeflag, boundarychar
# they each correspond with the property of the same name. Of them
# 'designsize' has the default value 10, but it can be cancelled
# by 'designsize' key with the explicit undef value.
# If $fd is not undef, it specifies FONTDIMEN list: if $fd is an
# array ref it is seen as pl-list of FONTDIMEN and placed inside
# the output pl-list; if $fd is a hash ref then pl_fontdimen($fd)
# is placed instead.
# Note: currently these three functions give the same result.

##<*> pl_header($in, $fd)
sub pl_header
{ return pl_header_gen($_[0], $_[1], 0); }

##<*> pl_header_vpl($in, $fd)
sub pl_header_vpl
{ return pl_header_gen($_[0], $_[1], 8); }

##<*> pl_header_opl($in, $fd, $swl1)
sub pl_header_opl
{ return pl_header_gen($_[0], $_[1], ($_[2]) ? 2 : 1); }

## pl_header_gen($in, $fd, $sw)
# Subcontractor for the above three pl_header_* functions.
sub pl_header_gen
{
  my ($in, $fd, $sw) = @_; my ($t, $pe, $dsiz, $ol, @pl);
  if (exists $in->{ofmlevel}) { $ol = $in->{ofmlevel}; }
  elsif ($sw == 1 || $sw == 2) { $ol = $sw - 1; }
  $dsiz = (exists $in->{designsize}) ? $in->{designsize} : 10;
  if (defined $ol) {
    $pe = pl_cook(['OFMLEVEL', 'H', $ol]);
    pl_set_value($pe, 1, $ol); push(@pl, $pe);
  }
  if (defined $in->{direction}) {
    push(@pl, ['DIRECTION', $in->{direction}]);
  }
  if (defined $in->{family}) {
    push(@pl, ['FAMILY', $in->{family}]);
  }
  if (defined $in->{vtitle}) {
    push(@pl, ['VTITLE', $in->{vtitle}]);
  }
  if (defined $in->{face}) {
    $pe = pl_cook(['FACE', 'F', 0]);
    pl_set_value($pe, 1, $in->{face}); push(@pl, $pe);
  }
  if (defined $in->{codingscheme}) {
    push(@pl, ['CODINGSCHEME', $in->{codingscheme}]);
  }
  if (defined $in->{designunits}) {
    $pe = pl_cook(['DESIGNUNITS', 'R', 0]);
    pl_set_real($pe, 1, $in->{designunits}); push(@pl, $pe);
  }
  if (defined $dsiz) {
    $pe = pl_cook(['DESIGNSIZE', 'R', 0]);
    pl_set_real($pe, 1, $dsiz); push(@pl, $pe);
  }
  if (defined $in->{checksum}) {
    $pe = pl_cook(['CHECKSUM', 'O', 0]);
    pl_set_value($pe, 1, $in->{checksum}); push(@pl, $pe);
  }
  if (defined $in->{sevenbitsafeflag}) {
    push(@pl, ['SEVENBITSAFEFLAG', $in->{sevenbitsafeflag}]);
  }
  if (ref $fd eq 'ARRAY') {
    push(@pl, $fd);
  }
  if (ref $fd eq 'HASH') {
    push(@pl, pl_fontdimen($fd));
  }
  if (defined $in->{boundarychar}) {
    $pe = pl_cook(['BOUNDARYCHAR', 'C', 0]);
    pl_set_value($pe, 1, $in->{boundarychar}); push(@pl, $pe);
  }
  return \@pl;
}

##<*> pl_fontdimen($in)
# Generates a FONTDIMEN list. $in is a hash ref with the following
# effective keys:
#   slant, space, stretch, shrink, xheight, quad, extraspace;
# they each correspond with the property of the same name. All
# of them have a default value.
sub pl_fontdimen
{
  my ($in) = @_; my ($q, $t, $pl);
  (defined $in) or $in = { };
  $pl = pl_cook(['FONTDIMEN',
         ['SLANT', 'R', 0], ['SPACE', 'R', 0], 
         ['STRETCH', 'R', 0], ['SHRINK', 'R', 0], 
         ['XHEIGHT', 'R', 0], ['QUAD', 'R', 0], 
         ['EXTRASPACE', 'R', 0]]);
  $q = $in->{quad}; (defined $q) or $q = 1;
  pl_set_real($pl->[6], 1, $q);
  $t = $in->{slant}; (defined $t) or $t = 0;
  pl_set_real($pl->[1], 1, $t);
  $t = $in->{space}; (defined $t) or $t = $q / 3;
  pl_set_real($pl->[2], 1, $t);
  $t = $in->{stretch}; (defined $t) or $t = $q / 6;
  pl_set_real($pl->[3], 1, $t);
  $t = $in->{shrink}; (defined $t) or $t = $q / 9;
  pl_set_real($pl->[4], 1, $t);
  $t = $in->{xheight}; (defined $t) or $t = $q / 2;
  pl_set_real($pl->[5], 1, $t);
  $t = $in->{extraspace}; (defined $t) or $t = $q / 9;
  pl_set_real($pl->[7], 1, $t);
  return $pl;
}

##<*> pl_fontdimen_jpl($in)
# Generates a FONTDIMEN list of JPL. Here $in is the same as in
# pl_fontdimen, except that two additional keys 'extrastretch'
# and 'extrashrink' are used and default setting is different.
sub pl_fontdimen_jpl
{
  my ($in) = @_; my ($q, $t, $pl);
  (defined $in) or $in = { };
  $pl = pl_cook(['FONTDIMEN',
         ['SLANT', 'R', 0], ['SPACE', 'R', 0], 
         ['STRETCH', 'R', 0], ['SHRINK', 'R', 0], 
         ['XHEIGHT', 'R', 0], ['QUAD', 'R', 0], 
         ['EXTRASPACE', 'R', 0], ['EXTRASTRETCH', 'R', 0], 
         ['EXTRASHRINK', 'R', 0]]);
  $q = $in->{quad}; (defined $q) or $q = 1;
  pl_set_real($pl->[6], 1, $q);
  $t = $in->{slant}; (defined $t) or $t = 0;
  pl_set_real($pl->[1], 1, $t);
  $t = $in->{space}; (defined $t) or $t = 0;
  pl_set_real($pl->[2], 1, $t);
  $t = $in->{stretch}; (defined $t) or $t = $q / 10;
  pl_set_real($pl->[3], 1, $t);
  $t = $in->{shrink}; (defined $t) or $t = 0;
  pl_set_real($pl->[4], 1, $t);
  $t = $in->{xheight}; (defined $t) or $t = $q;
  pl_set_real($pl->[5], 1, $t);
  $t = $in->{extraspace}; (defined $t) or $t = $q / 4;
  pl_set_real($pl->[7], 1, $t);
  $t = $in->{extrastretch}; (defined $t) or $t = $q / 5;
  pl_set_real($pl->[8], 1, $t);
  $t = $in->{extrashrink}; (defined $t) or $t = $q / 8;
  pl_set_real($pl->[9], 1, $t);
  return $pl;
}

# for pl_fonrdimen_?pl_rmt
our @pl_keys_quad_u_jpl = (
  0x4E00, 0x3042, 0x306E, 0xFF2D, 0x004D, 0x2014
);
our @pl_keys_quad_u_opl = (
  0x2001, 0x2003, 0x004D, 0x2014
);
our @pl_keys_space_u_opl = (
  0x0020, 0x00A0
);

##<*> pl_fontdimen_opl_rmt($rmt)
# Generates a FONTDIMEN list of OPL with values estimated from
# the glyph metric $rmt.
sub pl_fontdimen_opl_rmt
{
  my ($rmt) = @_; my ($t, $in);
  $in = { };
  (defined($t = pl_key_width($rmt, \@pl_keys_quad_u_opl)))
    and $in->{quad} = $t;
  (defined($t = pl_key_width($rmt, \@pl_keys_space_u_opl)))
    and $in->{space} = $t;
  return pl_fontdimen($in);
}

##<*> pl_fontdimen_jpl_rmt($rmt)
# Generates a FONTDIMEN list of JPL with values estimated from
# the glyph metric $rmt.
sub pl_fontdimen_jpl_rmt
{
  my ($rmt) = @_; my ($t, $in);
  $in = { };
  (defined($t = pl_key_width($rmt, \@pl_keys_quad_u_jpl)))
    and $in->{quad} = $t;
  return pl_fontdimen($in);
}

## pl_key_width($rmt)
# Subcontractor of pl_fontdimen_?pl_rmt.
sub pl_key_width
{
  my ($rmt, $keys) = @_; my ($t, %hsh);
  foreach (@$rmt) { $hsh{$_->[0]} = $_; }
  foreach (@$keys) {
    (defined($t = $hsh{$_})) or next;
    return $t->[1];
  }
  return;
}

## pl_cook($pl)
# Cooks a pl-list $pl and returns it. (Sometimes more convenient
# than pl_cook_list.)
sub pl_cook
{
  my ($pl) = @_;
  (pl_cook_list($pl)) or return;
  return $pl;
}

# for pl_get_metric
my %pl_char_part_pos = (
  CHARWD => 1, CHARHT => 2, CHARDP => 3, CHARIC => 4
);

##<*> pl_get_metric($pl)
# Reads the metric data from the CHARACTER set of pl-struct $pl
# and converts them to a raw metric array.
sub pl_get_metric
{
  my ($pl) = @_; my ($t, $p, $pe, $pe2, $ent, @rmt);
  foreach $pe (@$pl) {
    if ($pe->[0] eq 'CHARACTER') {
      (defined($t = pl_value($pe, 1))) or return;
      $ent = [ $t ];
      foreach $pe2 (@{$pe}[3 .. $#$pe]) {
        if (defined($p = $pl_char_part_pos{$pe2->[0]})) {
          (defined($t = pl_value($pe2, 1))) or return;
          $ent->[$p] = $t / TU;
        }
      }
      push(@rmt, $ent);
    }
  }
  return \@rmt;
}

# for pl_char_part
my @pl_char_part_name = qw( * CHARWD CHARHT CHARDP CHARIC );

##<*> pl_char_part
# Converts a raw metric array to an array of CHARACTER lists.
# Partial inverse of pl_get_metric
sub pl_char_part
{
  my ($rmt) = @_; my ($ent, $pe, $pe2, $pl);
  $pl = [ ];
  foreach $ent (@$rmt) {
    $pe = pl_cook(['CHARACTER', 'C', 0]);
    pl_set_value($pe, 1, $ent->[0]);
    foreach (1 .. 4) {
      if (defined($ent->[$_])) {
        $pe2 = pl_cook([$pl_char_part_name[$_], 'R', 0]);
        pl_set_real($pe2, 1, $ent->[$_]); push(@$pe, $pe2);
      }
    }
    push(@$pl, $pe);
  }
  return $pl;
}

##<*> pl_char_part_jpl($imt, $vmt, $swu)
# ...
sub pl_char_part_jpl
{
  my ($imt, $vmt, $swu) = @_; my ($t, $cc, $ti, $jp);
  my (@cit, $pl, $pe, @u);
  $jp = ($swu) ? 'U' : 'J';
  foreach $cc (keys %$imt) {
    push(@{$cit[$imt->{$cc}]}, $cc);
  }
  foreach $ti (1 .. $#cit) {
    @u = map { sprintf("%s%04X", $jp, $_) }
             (sort { $a <=> $b } (@{$cit[$ti]}));
    $pe = pl_cook(['CHARSINTYPE', 'D', 0, @u]);
    pl_set_value($pe, 1, $ti); push(@$pl, $pe);
  }
  foreach $ti (0 .. $#$vmt) {
    $pe = pl_cook(['TYPE', 'D', 0, ['CHARWD', 'R', 0],
                   ['CHARHT', 'R', 0], ['CHARDP', 'R', 0]]);
    pl_set_value($pe, 1, $ti);
    pl_set_real($pe->[3], 1, $vmt->[$ti][0]);
    pl_set_real($pe->[4], 1, $vmt->[$ti][1]);
    pl_set_real($pe->[5], 1, $vmt->[$ti][2]);
    push(@$pl, $pe);
  }
  return $pl;
}


######## 'jcode' section ########

our $jcode_ex = ($ENV{OS} =~ /Windows/) ? 'shiftjis' : 'euc-jp';
our $jcode_in = 'jis0208-raw';

my %jcode_ex_sym = ( jis => 'iso-2022-jp', euc => 'euc-jp',
                     sjis => 'shiftjis', utf8 => 'UTF-8' );
my %jcode_in_sym = ( jis => 'jis0208-raw', unicode => 'UTF-16BE' );

##<*> jcode_set($xjc, $ijc)
# Declares external [internal] Japanese code to be $xjc [$ijc],
# which is a key of %jcode_ex_sym [%jcode_in_sym].
sub jcode_set
{
  my ($xjc, $ijc) = @_;
  if (defined $xjc) {
    (defined($jcode_ex = $jcode_ex_sym{$xjc})) or return;
  }
  if (defined $ijc) {
    (defined($jcode_in = $jcode_in_sym{$ijc})) or return;
  }
  return 1;
}

##<*> jcode_chr($cod)
# Converts a code value of the internal code to a string containing
# the character encoded in the external code.
sub jcode_chr
{
  my ($cod) = @_; my ($xs, $is);
  (0 <= $cod && $cod <= 0xFFFF) or return;
  $xs = chr($cod >> 8) . chr($cod & 0xFF);
  eval {
    $is = decode($jcode_in, $xs, Encode::FB_CROAK);
    $xs = encode($jcode_ex, $is, Encode::FB_CROAK);
  };
  ($@ eq '' && length($is) == 1) or return;
  return $xs;
}

##<*> jcode_ord($xs)
# Inverse of jcode_chr.
sub jcode_ord
{
  my ($xs) = @_; my ($is, $cod);
  eval {
    $is = decode($jcode_ex, $xs, Encode::FB_CROAK);
    $xs = encode($jcode_in, $is, Encode::FB_CROAK);
  };
  ($@ eq '' && length($is) == 1 && $xs =~ /^(.)(.)$/) or return;
  return ord($1) << 8 | ord($2);
}

######## 'ps' section ########

#### ps_parse0: shallow parser
# The 'shallow parser' converts a whole PS text into a list of
# (pseudo)tokens (each of which is a simple scalar value):
# where '[', ']', '<<', '>>' are treated as a separate token
# and represented as strings read as itself; and other tokens
# are represented as follows.
#   number:   +123.0     -> number 123
#   operator: abc        -> string _abc
#   name:     /abc       -> string /abc
#   literal   (abc)      -> string 'abc
#   hexstring <616263>   -> string "abc

sub ps_parse0
{
  my ($txt) = @_; my ($p, $tok, @toks);
  while ($txt =~ m/\G((\s+)|<<|>>|(\%)|(<)|(\()| #)
                      (\/?[^<>()\/\[\]\{\}\%\s]+)|.)/gx) {
    $p = pos($txt);
    if (defined $2) { next; } # spaces
    elsif (defined $3) { # '% comment'
      $txt =~ m/\G.*$/gm; next; 
    } elsif (defined $4) { # <hexstring>
      ($tok, pos($txt)) = ps_parse0_hexstr($txt, $p);
      (defined $tok) or return ps_parse0_error($txt, $p - 1);
    } elsif (defined $5) { # (literalstring)
      ($tok, pos($txt)) = ps_parse0_litstr($txt, $p);
      (defined $tok) or return ps_parse0_error($txt, $p - 1);
    } elsif (defined $6) {
      $tok = $6;
      if ($tok =~ m/^[+-]?\d+(?:\.\d*)?$/) { $tok += 0; }
      elsif ($tok !~ m|^/|) { $tok = "_$tok"; }
    } else { $tok = $1; }
    push(@toks, $tok);
  }
  return \@toks;
}
sub ps_parse0_hexstr
{
  my ($txt, $p) = @_; my ($t);
  pos($txt) = $p; $txt =~ m/([\d\sA-Fa-f]*)/g; $t = $1;
  ($txt =~ m/\G>/g) or return;
  $t =~ s/\s+//g; $t = pack("H*", $t);
  return ("\"$t", pos($txt));
}
sub ps_parse0_litstr
{
  my ($txt, $p) = @_; my ($t);
  pos($txt) = $p;
  ($txt =~ m/(.*?)\)/gs) or return; $t = $1;
  return ("\'$t", pos($txt));
}

sub ps_s { return substr($_[0], 1); }
sub ps_form
{
  my ($tok) = @_;
  return ($tok =~ m/^[\"\']/) ? "<str>" :
         ($tok =~ m/^_/) ? ps_s($tok) : $tok;
}


sub ps_parse0_error
{
  my ($txt, $p) = @_; my ($t);
  $t = substr($txt, $p, 10);
  return error("syntax error in PS code (near $t...)");
}

######## 'enc' section ########

# The data structure for encoding vectors is merely an array ref
# of glyph names. Usually this array has 256 elements but can be
# otherwise. The name '.notdef' is represented as undef.
# In addition, this module provides a registry of encodings, which
# is a hash that maps an enc name to an enc array.

##<*> enc_parse($txt)
# Converts a text $txt in the form of '.enc' files to an enc array.
# In list context it returns ($key, $enc) where $key is the name of
# encoding and $enc is the array, while in scalar context it returns
# $enc only.
sub enc_parse
{
  my ($txt, $sw) = @_; my ($dat, $key, @enc);
  L1:{
    (defined($dat = ps_parse0($txt))) or last;
    ($dat->[1] eq '[' && $dat->[-2] eq ']' && $dat->[-1] eq '_def' &&
     $dat->[0] =~ m|^/(.*)|) or last;
    $key = $1;
    foreach (2 .. $#$dat - 2) {
      ($dat->[$_] =~ m|^/(.*)|) or last L1;
      push(@enc, ($1 eq '.notdef' && !$sw) ? undef : $1);
    }
    enc_register($key, \@enc);
    return (wantarray) ? ($key, \@enc) : \@enc;
  }
  return error("syntax error in enc data");
}

# for enc_form
use constant {
  ENCF_NORMAL => 0, ENCF_ANNOT => 1, ENCF_COMPACT => 2,
  ENCF_256 => 256
};
our $enc_form_indwd = 2;     # indent width
our $enc_form_annwd = 24;    # column where comments stay
our $enc_form_linwd = 72;    # maximum line length

##<*> enc_form($key, $enc, $mod)
# Converts enc array $enc to the text of .enc file. Here $key is
# the name the encoding vector will be defined with, and $mod 
# specifies the output form and is OR-value of the following:
#   ENCF_NORMAL:     normal, one line per name
#   ENCF_ANNOT:      with comment like "% 0x18" per 8 lines
#   ENCF_COMPACT:    crams to a line as many names as possible
#   ENCF_256:        see as if the array has 256 entries
# If $enc is undef, array registed by name $key is used.
sub enc_form
{
  my ($key, $enc, $mod) = @_;
  my ($swsl, $k, $t, $l, $s, @lins, @enx);
  if (ref $enc ne 'ARRAY') {
    $mod = $enc;
    (defined($enc = enc_retrieve($key)))
      or return error("unknown enc name: $key");
  }
  $l = ($mod & 256) ? 255 : $#$enc; $mod = $mod & 15;
  foreach $k (0 .. $l) {
    $t = $enc->[$k]; $enx[$k] = (defined $t) ? "/$t" : '/.notdef';
  }
  @lins = ("/$key ["); $s = ' ' x $enc_form_indwd;
  if ($mod == ENCF_NORMAL) {
    foreach $k (@enx) { push(@lins, "$s$k"); }
  } elsif ($mod == ENCF_ANNOT) {
    foreach $k (0 .. $#enx) {
      $t = "$s" . $enx[$k];
      if ($k % 8 == 0) {
        $l = $enc_form_annwd - length($t); if ($l < 1) { $l = 1; }
        $t .= sprintf("%s%% 0x%02X", ' ' x $l, $k);
      }
      push(@lins, $t);
    }
  } elsif ($mod == ENCF_COMPACT) {
    $t = '';
    foreach $k (@enx) {
      if ($t eq '') { $t = $k; }
      elsif (length($t) + length($k) + 1 > $enc_form_linwd) {
        push(@lins, $t); $t = $k;
      } else { $t .= " $k"; }
    }
    push(@lins, $t);
  } else { return error("invalid mode given: $mod"); }
  push(@lins, "] def", '');
  return join("\n", @lins);
}

#### Procedures for the encoding registry

our %enc_registry;

##<*> enc_register($key, $enc)
# Register enc array $enc for name $key. If $enc is undef, then
# the entry for $key is deleted.
sub enc_register
{
  my ($key, $enc) = @_;
  if (defined $enc) { $enc_registry{$key} = $enc; }
  else { delete $enc_registry{$key}; }
}

##<*> enc_retrieve($key)
# Returns the enc array registered by name $key.
sub enc_retrieve
{
  return $enc_registry{$_[0]};
}

##<*> enc($key, $c)
# Returns the glyph name of code $c in the enc array registered
# by name $key.
sub enc
{
  return $enc_registry{$_[0]}[$_[1]];
}

######## 'kpse' section ########

our $kpsewhich = 'kpsewhich';
our $kpse_init_done;
our $kpse_delim;
our %kpse_format_alias = (
  cmap => 'cmap files',
);

##<*> kpse($fnam, $opt)
# Executes 'kpsewhich' for filename $fnam with option $opt.
# If $opt is a scalar, it means the value for 'format' option.
# If $opt is a hash ref, then the value for keys 'dpi', 'engine',
# 'mode', 'progname' and 'format' corresponds with the value of the
# option with same name and the boolean value for key 'mustexist'
# corresponds with existence of 'must-exist' option. For 'option'
# value, aliasing specified with %kpse_format_alias is done.
sub kpse
{
  my ($fnam, $opt) = @_; my ($res);
  ($kpse_init_done || kpse_init()) or return undef;
  $opt = kpse_parse_option($opt);
  if (ref $opt eq 'ARRAY') { return kpse_manual($fnam, $opt); }
  $res = `$kpsewhich $opt "$fnam"`; chomp($res);
  if (-f $res) { return $res; }
  else {                            # returns undef, not nothing
    error("kpse failed to find a file: $fnam"); return undef;
  }
}

##<*> kpse_init()
# Initializes the kpse section of this module.
sub kpse_init
{
  my ($res);
  (!defined $kpse_init_done)
    or return error("kpsewhich failure");
  if (($res = `$kpsewhich -show-path=tex`) eq '') {
    $kpse_init_done = 0;
    return error("kpsewhich failure");
  }
  if ($res =~ m/^\.\:/) { $kpse_delim = ':'; }
  elsif ($res =~ m/;/) { $kpse_delim = ';'; }
  else { $kpse_delim = ':'; }
  $kpse_init_done = 1;
  return 1;
}

## kpse_parse_option($opt)
# Subcontractor of kpse.
sub kpse_parse_option
{
  my ($opt) = @_; my ($o, $t, @copts);
  if ($opt eq '') { return; }
  elsif (ref $opt eq 'ARRAY') { return $opt; } # for future extension
  elsif (ref $opt eq 'HASH') {
    foreach $o (qw(dpi engine mode progname)) {
      if (exists $opt->{$o}) {
        push(@copts, "-$o=" . $opt->{$o});
      }
    }
    if ($opt->{mustexist}) { push(@copts, '-must-exist'); }
    $opt = $opt->{format};
  }
  if ($opt ne '') {
    if (defined($t = $kpse_format_alias{$opt})) { $opt = $t; }
    push(@copts, "-format=\"$opt\"");
  }
  return join(' ', @copts);
}

######## 'ttf' section ########

# * Data structure: raw metric
# A raw metric struct holds the most basic metric data of a font and
# is formed as follows: a raw metric is an array ref of character
# metric entries, which are of the following form:
#  [ <code>, <width>, <height>, <depth> ]
# here the unit of length is the designated size.

##<*> ttf_generate_opl($ffnt, $id)
# Read the metric of a TTF file $ffnt (in case of TTC index is given
# by $id) and generates the pl-struct describing ofm for the metric.
sub ttf_generate_opl
{
  my ($ffnt, $id) = @_; my ($t, $rmt, $fd, $pl, $chp);
  ($t = $ffnt) =~ s|.*/(.*)|$1|; 
  $t = { family => $t, codingscheme => 'UNICODE' };
  (defined($rmt = ttf_get_metric($ffnt, $id)) &&
   defined($fd = pl_fontdimen_opl_rmt($rmt)) &&
   defined($pl = pl_header_opl($t, $fd, 0)) &&
   defined($chp = pl_char_part($rmt))) or return;
  push(@$pl, @$chp);
  return $pl;
}

# for ttf_get_metric
  # This table designates the procedure to get raw metric of
  # a given type (known by extension) of font file.
our %ttf_get_metric_proc = (
 #ttf => \&ttf_get_metric_ttf2tfm,
  ttf => \&ttf_get_metric_xetex,
  otf => \&ttf_get_metric_xetex,
);

##<*> ttf_get_metric($fnam, $id)
sub ttf_get_metric
{
  my ($ffnt, $id) = @_; my ($rmt, $prc);
  L1:{
    if ($ffnt =~ m/\.ttf$/i) {
      $prc = $ttf_get_metric_proc{ttf};
      (defined($rmt = &$prc($ffnt))) or last L1;
    } elsif ($ffnt =~ m/\.ttc$/i) {
      $prc = $ttf_get_metric_proc{ttf};
      (defined($rmt = &$prc($ffnt, $id))) or last L1;
    } elsif ($ffnt =~ m/\.(\w{3,4})$/i) {
      $prc = $ttf_get_metric_proc{lc($1)};
      (defined($rmt = &$prc($ffnt, $id))) or last L1;
    } else { last L1; }
    return $rmt;
  }
  return error("failed to get font metric: $ffnt");
}

## ttf_get_metric_ttf2tfm($fnam, $id)
sub ttf_get_metric_ttf2tfm
{
  my ($ffnt, $id) = @_; my ($t, $tb, $cout, $rmt, @tfms);
  $tb = get_temp_name();
  @tfms = map { sprintf("%s-%02x.tfm", $tb, $_) } (0 .. 255);
  unlink(@tfms);
  $t = (defined $id) ? "-f $id" : '';
  $cout = `ttf2tfm "$ffnt" $t $tb-\@Unicode\@`;
  (unlink(@tfms) > 0) or return;
  return ttf_parse_out_ttf2tfm($cout);
}

# for ttf_parse_out_ttf2tfm
our $ttf_ttf2tfm_scale = 1000;
  # Is that always 1000?

## ttf_parse_out_ttf2tfm(...)
# Subcontractor of ttf_get_metric_ttf2tfm.
sub ttf_parse_out_ttf2tfm
{
  my ($txt) = @_; my ($scl, $lin, $stg, @fs, @rmt);
  $scl = $ttf_ttf2tfm_scale;
  foreach $lin (split(/\n/, $txt)) {
    if ($lin eq '') { $stg = 0; }
    elsif ($stg == 0 && $lin =~ m/^\s*Glyph/) { $stg = 1; }
    elsif ($stg == 1 && $lin =~ m/^\s*-----/) { $stg = 2; }
    elsif ($stg == 2 && (@fs = $lin =~
            m/^\s*(\d+)\s+(\w+)\s+(\S+)\s+(-?\d+)
              \s+(-?\d+),\s*(-?\d+)\s*--\s*(-?\d+),\s*(-?\d+)/x)) {
      push(@rmt, [ hex($fs[1]), $fs[3] / $scl, $fs[7] / $scl,
                   -$fs[5] / $scl ]);
    }
  }
  return \@rmt;
}

## ttf_get_metric_xetex(...)
# Dispatcher of ttf_get_metric for OpenType fonts.
sub ttf_get_metric_xetex
{
  my ($ffnt) = @_; my ($t, $ftb, $cout);
  $ftb = get_temp_name();
  $t = <<"END1" . <<'END2';
\\def\\fn{$ffnt}
END1
\newcount\cc \newbox\cb
\font\fs="[\fn]"\relax\fs
\immediate\write-1{>>>Start}%
\cc=0 \loop \ifnum\XeTeXcharglyph\cc>0
    \setbox\cb\hbox{\char\cc}%
    \immediate\write-1{>>\the\cc/\the\wd\cb/\the\ht\cb/\the\dp\cb}%
  \fi
\advance\cc1 \ifnum\cc<"10000 \repeat
\immediate\write-1{>>>End}%
\bye
END2
  (write_whole_file("$ftb.tex", $t)) or return;
  $cout = `xetex -halt-on-error -no-pdf $ftb`;
  (defined($t = read_whole_file("$ftb.log"))) or return;
  unlink("$ftb.tex", "$ftb.log", "$ftb.xdv");
  return ttf_parse_out_xetex($t);
}

## ttf_parse_out_xetex(...)
# Subcontractor of ttf_get_metric_xetex.
sub ttf_parse_out_xetex
{
  my ($txt) = @_; my ($stg, $lin, @fs, @rmt);
  $stg = 0;
  foreach $lin (split(/\n/, $txt)) {
    if ($stg == 0) { 
      if ($lin eq '>>>Start') { $stg = 1; }
    } elsif ($stg == 1) {
      if (@fs = $lin =~
          m|^>>(\d+)/(-?[\.\d]+)pt/(-?[\.\d]+)pt/(-?[\.\d]+)pt$|) {
        push(@rmt, [ $fs[0], map { sprintf("%.5f", $_ / 10) - 0 }
                                 (@fs[1, 2, 3]) ]);
      } elsif ($lin eq '>>>End') { $stg = 2; }
      else { last; }
    } else { last; }
  }
  if ($stg != 2) { return; }
  return \@rmt;
}

######## 'tfm' section ########

##<*> tfm_reduce_jpl($rmt, $d)
# Constructs a JFM metric, making the best effort to minimize the
# maximum height/depth error under the condition the maximum width
# error must be at most $d.
sub tfm_reduce_jpl
{
  my ($rmt, $dw) = @_; my ($imt, $vmt, $d, $dl, $dh);
  ($imt, $vmt) = tfm_try_reduce_jpl($rmt, $dw, 0);
  if (defined $vmt) { return (0, $imt, $vmt); }
  for ($d = 1; $d <= 2048; $d *= 2) {
    ($imt, $vmt) = tfm_try_reduce_jpl($rmt, $dw, $d);
    if (defined $vmt) { last; }
  }
  ($d <= 2048) or return;
  $dl = 0; $dh = $d;
  while ($dh - $dl > 9e-7) {  # 9e-7 < 2**-20
    $d = ($dl + $dh) / 2;
    ($imt, $vmt) = tfm_try_reduce_jpl($rmt, $dw, $d);
    if (defined $vmt) { $dh = $d; }
    else { $dl = $d; }
  }
  $d = $dh; ($imt, $vmt) = tfm_try_reduce_jpl($rmt, $dw, $d);
  (defined $vmt) or return fatal("tfm_reduce_jpl_x");
  return ($dw, $d, $imt, $vmt);
}

##<*> tfm_reduce_jpl_x($rmt, $bf)
# Constructs a JFM metric, making the best effort to minimize the
# value d that satisfies the condition the maximum error of width,
# height, and depth must be at most d/$bf, d, and d.
sub tfm_reduce_jpl_x
{
  my ($rmt, $bf) = @_; my ($imt, $vmt, $d, $dl, $dh);
  if (!defined $bf) { $bf = 20; }
  ($imt, $vmt) = tfm_try_reduce_jpl($rmt, 0, 0);
  if (defined $vmt) { return (0, $imt, $vmt); }
  for ($d = 1; $d <= 2048; $d *= 2) {
    ($imt, $vmt) = tfm_try_reduce_jpl($rmt, $d / $bf, $d);
    if (defined $vmt) { last; }
  }
  ($d <= 2048) or return;
  $dl = 0; $dh = $d;
  while ($dh - $dl > 1e-7) {
    $d = ($dl + $dh) / 2;
    ($imt, $vmt) = tfm_try_reduce_jpl($rmt, $d / $bf, $d);
    if (defined $vmt) { $dh = $d; }
    else { $dl = $d; }
  }
  $d = $dh; ($imt, $vmt) = tfm_try_reduce_jpl($rmt, $d / $bf, $d);
  (defined $vmt) or return fatal("tfm_reduce_jpl_x");
  return ($d / $bf, $d, $imt, $vmt);
}

## tfm_try_reduce_jpl($rmt, $d1, $d2)
# Subcontractor of tfm_reduce_jpl_x. (Here $d1 is the error limit of
# widths and $d2 is of heights and depths.
sub tfm_try_reduce_jpl
{
  my ($rmt,  $d1, $d2) = @_; my ($j, $t, $e, @u, %cnt);
  my (@ovs, @is, @vs, %imt, @vmt);
  @u = ([], [$d1, 0], [$d2, 1], [$d2, 1]);
  foreach $t (1 .. 3) {
    @{$ovs[$t]} = map { $_->[$t] } (@$rmt);
    ($is[$t], $vs[$t]) = tfm_set_indices_j($ovs[$t], @{$u[$t]});
    (scalar(@{$vs[$t]}) <= 256) or return;
  }
  foreach $j (0 .. $#$rmt) {
    @u = map { pl_form_real($vs[$_][$is[$_][$j]]) } (1 .. 3);
    $imt{$rmt->[$j][0]} = $t = join('/', @u); $cnt{$t} += 1;
  }
  @vmt = keys %cnt; (scalar(@vmt) <= 256) or return;
  @vmt = sort { $cnt{$b} <=> $cnt{$a} } (@vmt);
  foreach (0 .. $#vmt) { $cnt{$vmt[$_]} = $_; }
  foreach (keys %imt) {  $imt{$_} = $cnt{$imt{$_}}; }
  @vmt = map { [split(m|/|, $_)] } (@vmt);
  return (\%imt, \@vmt);
}

###<*> tfm_reduce($rmt)
sub tfm_reduce
{
  my ($rmt) = @_; my ($t, $e, %cnt);
  my (@wds, @hts, @dps, $iht, $vht, $idp, $vdp, %imt, @vmt);
  (@$rmt) or return error("metric data are empty");
  @wds = map { $_->[1] } (@$rmt);
  (($t = tfm_count_kind(\@wds)) <= 255)
    or return error("too many kinds ($t > 255) of widths");
  @hts = map { $_->[2] } (@$rmt);
  ($iht, $vht) = tfm_set_indices(\@hts, 16);
  @dps = map { $_->[3] } (@$rmt);
  ($idp, $vdp) = tfm_set_indices(\@dps, 16);
  foreach $e (@$rmt) {
    $t = join('/', $e->[1], $vht->[$iht->[$e->[2]]],
                   $vdp->[$idp->[$e->[3]]]);
    $imt{$e->[0]} = $t; $cnt{$t} += 1;
  }
  @vmt = sort { $cnt{$b} <=> $cnt{$a} } (keys %cnt);
  foreach (0 .. $#vmt) { $cnt{$vmt[$_]} = $_; }
  foreach (keys %imt) {  $imt{$_} = $cnt{$imt{$_}}; }
  foreach $t (0 .. $#vmt) {
    $vmt[$t] = [ (split(m|/|, $vmt[$t])) ];
  }
  return (\%imt, \@vmt);
}

## tfm_count_kind($ary)
# Returns the number of distinct elements in an array $ary.
sub tfm_count_kind
{
  my ($ary) = @_; my (%hsh, @a2);
  foreach (@$ary) { $hsh{$_} += 1; }
  @a2 = keys %hsh; return scalar(@a2);
}

## tfm_set_indices($ary, $m)
# Reduces a set $ary of reals to ($m - 1) elements using the
# same rounding algorithm as in pltotf (here $m is decremented
# because zero comes into the list later). It returns ($idx, $val)
# where $val is the reduced set and $idx is the index array
# for the original array $ary.
sub tfm_set_indices
{
  my ($ary, $m) = @_; my ($v, $v1, $d, @ary2, @idx, @val, %hsh);
  foreach $v (@$ary) { $hsh{$v} = 1; }
  delete $hsh{0}; @ary2 = sort { $a <=> $b } (keys %hsh);
  --$m; $d = tfm_shorten(\@ary2, $m);
  @val = (0); %hsh = (0, 0); $m = 0; $v1 = -2 * HUGE;
  foreach $v (@ary2) {
    if ($v > $v1 + $d) {
      $hsh{$v} = ++$m; $val[$m] = $v1 = $v;
    } else {
      $val[$m] = ($v1 + $v) / 2; $hsh{$v} = $m;
    }
  }
  @idx = map { $hsh{$_} } (@$ary);
  return (\@idx, \@val, $d);
}

## tfm_set_indices_j($ary, $d, $sw)
# Similar as tfm_set_indices, but here $d is the maximum permitted
# value of round error. If $sw is 0/1/2 ,then values in a range
# is rounded to mid-value/lower-limit/upper-limit of the range.
sub tfm_set_indices_j
{
  my ($ary, $d, $sw) = @_;
  my ($v, $v1, $m, @ary2, @idx, @val, %hsh);
  foreach $v (@$ary) { $hsh{$v} = 1; }
  delete $hsh{0}; @ary2 = sort { $a <=> $b } (keys %hsh);
  @val = (0); %hsh = (0, 0); $m = 0; $v1 = -2 * HUGE;
  foreach $v (@ary2) {
    if ($v > $v1 + $d) {
      $hsh{$v} = ++$m; $val[$m] = $v1 = $v;
    } else {
      $val[$m] = ($sw == 1) ? $v1 : ($sw == 2) ? $v :
                 (($v1 + $v) / 2);
      $hsh{$v} = $m;
    }
  }
  @idx = map { $hsh{$_} } (@$ary);
  return (\@idx, \@val);
}

## tfm_min_cover($ary, $d)
# Subcontractor of tfm_indices.
sub tfm_min_cover
{
  my ($ary, $d) = @_; my ($v, $d1, $v1, $dnxt, $m);
  $m = 0; $dnxt = HUGE; $v1 = -2 * HUGE;
  foreach $v (@$ary) {
    if ($v > $v1 + $d) {
      $d1 = $v - $v1; $v1 = $v; ++$m;
      if ($d1 < $dnxt) { $dnxt = $d1; }
    }
  }
  return ($m, $dnxt);
}

## tfm_shorten($ary, $m)
# Subcontractor of tfm_indices.
sub tfm_shorten
{
  my ($ary, $m) = @_; my ($k, $d, $d1);
  ($k, $d) = tfm_min_cover($ary, 0);
  if ($k <= $m) { return 0; }
  while ($k > $m) { $d += $d; ($k) = tfm_min_cover($ary, $d); }
  $d /= 2; ($k, $d1) = tfm_min_cover($ary, $d);
  while ($k > $m) { $d = $d1; ($k, $d1) = tfm_min_cover($ary, $d); }
  return $d;
}

##<*> tfm_reduced_metric($imt, $vmt)
sub tfm_reduced_metric
{
  my ($imt, $vmt) = @_; my ($cc, @ccs, $t, @rmt);
  @ccs = sort { $a <=> $b } (keys %$imt);
  foreach $cc (@ccs) {
    $t = $vmt->[$imt->{$cc}];
    push(@rmt, [$cc, @$t]);
  }
  return \@rmt;
}

##<*> tfm_map_rmt($rmt, $map)
# Remaps the character code used in a raw metric $rmt by applying
# $map backwards.
sub tfm_map_rmt
{
  my ($rmt, $map) = @_; my ($e, $t, @rmt2, %hsh);
  (defined($map = arraymap($map))) or return;
  foreach (@$rmt) { $hsh{$_->[0]} = $_; }
  foreach $e (@$map) {
    (defined($t = $hsh{$e->[1]})) or next;
    push(@rmt2, [$e->[0], @{$t}[1 .. $#$t]]);
  }
  return \@rmt2;
}


######## 'vf' section ########

##<*> vf_parse($dat, $swdh)
# Converts a (binary) VF data $dat to a pl-struct describing
# VPL0 data. If something invalid is found in DVI code and $swdh
# is true, then DVI is written with a DIRECTHEX entry.
sub vf_parse
{
  my ($dat, $swdh) = @_;
  my ($t, $u, @fs, $pos, $pl, $pe, $stg);
  (length($dat) >= 3) or return vf_synerror("in preamble");
  @fs = unpack("CCC/a*NN", $dat); $pos = length($fs[2]) + 11;
  ($#fs == 4 && $fs[0] == 247 && $fs[1] == 202)
    or return vf_synerror("in preamble");
  $pl = pl_header_vpl({ vtitle => $fs[2], checksum => $fs[3],
                        designsize => $fs[4] / TU });
  for (;;) {
    $t = ord(substr($dat, $pos, 1));
    if ($stg <= 2 && 0 <= $t && $t <= 241) { # short_charN
      @fs = unpack("CCa3a$t", substr($dat, $pos)); $pos += $t + 5;
      ($#fs == 3 && length($fs[3]) == $t)
        or return vf_synerror("premature end");
      $pe = pl_cook(['CHARACTER', 'C', 0,
                     ['CHARWD', 'R', 0], undef]);
      if (defined($t = vf_dvi_parse($fs[3]))) { $pe->[4] = $t; }
      elsif (!$swdh) {
        return vf_synerror("illegal dvi code (char $fs[1])");
      } else { $pe->[4] = vf_dvi_dumb_parse($u); }
      pl_set_value($pe->[3], 1, unpack_num($fs[2])); # (unsigned)
      pl_set_value($pe, 1, $fs[1]);
      $stg = 2; push(@$pl, $pe);
    } elsif ($stg <= 2 && $t == 242) { # long_char
      @fs = unpack("CNNN", substr($dat, $pos)); $pos += 13;
      $u = substr($dat, $pos, $fs[1]); $pos += $fs[1];
      $pe = pl_cook(['CHARACTER', 'C', 0,
                     ['CHARWD', 'R', 0], undef]);
      if (defined($t = vf_dvi_parse($u))) { $pe->[4] = $t; }
      elsif (!$swdh) {
        return vf_synerror("illegal dvi code (char $fs[2])");
      } else { $pe->[4] = vf_dvi_dumb_parse($u); }
      pl_set_value($pe->[3], 1, signed($fs[3]));
      pl_set_value($pe, 1, $fs[2]);
      $stg = 2; push(@$pl, $pe);
    } elsif ($stg <= 1 && 243 <= $t && $t <= 246) { # fnt_defN
      $t -= 242;
      @fs = unpack("Ca${t}NNNCC", substr($dat, $pos)); $pos += $t + 15;
      ($#fs == 6) or return vf_synerror("premature end");;
      $t = $fs[5] + $fs[6]; $u = substr($dat, $pos, $t); $pos += $t;
      (length($u) == $t) or return vf_synerror("premature end");;
      $fs[6] = substr($u, $fs[5]); $fs[5] = substr($u, 0, $fs[5]);
      $pe = pl_cook(['MAPFONT', 'D', 0, ['FONTAREA', 0],
                     ['FONTNAME', 0],, ['FONTCHECKSUM', 'O', 0],
                     ['FONTAT', 'R', 0], ['FONTDSIZE', 'R', 0]]);
      $pe->[3][1] = $fs[5]; $pe->[4][1] = $fs[6];
      pl_set_value($pe->[5], 1, $fs[2]);
      pl_set_value($pe->[6], 1, $fs[3]);
      pl_set_value($pe->[7], 1, $fs[4]);
      pl_set_value($pe, 1, unpack_num($fs[1]));
      if ($fs[5] eq '') { splice(@$pe, 3, 1); }
      $stg = 1; push(@$pl, $pe);
    } elsif ($stg == 2 && $t == 248) { # post
      (($u = substr($dat, $pos, $t)) =~ /^\xf8+$/)
        or return vf_synerror("in postamble");
      last;
    } else { return vf_synerror("unexpected byte $t"); }
  }
  return $pl;
}

## vf_dvi_parse($dat)
# Subcontractor of vf_parse.
sub vf_dvi_parse
{
  my ($dat) = @_;
  my ($t, $u, @fs, $pos, $pl, $pe, $stk, $stg);
  $pl = ['MAP']; $stk = [{}];
  for ($pos = 0; $pos < length($dat); ) {
    $t = ord(substr($dat, $pos, 1));
    if (0 <= $t && $t <= 127) { # set_charN
      $pe = pl_cook(['SETCHAR', 'C', 0]); $pos += 1;
      pl_set_value($pe, 1, $t); push(@$pl, $pe);
    } elsif (128 <= $t && $t <= 131) { # setN
      $t -= 127; @fs = unpack("Ca$t", substr($dat, $pos));
      $pos += $t + 1; ($#fs == 1) or return;
      $pe = pl_cook(['SETCHAR', 'C', 0]);
      pl_set_value($pe, 1, unpack_num($fs[1])); push(@$pl, $pe);
    } elsif ($t == 132) { # setrule
      @fs = unpack("CNN", substr($dat, $pos));
      $pos += 9; ($#fs == 2) or return;
      $pe = pl_cook(['SETRULE', 'R', 0, 'R', 0]);
      pl_set_value($pe, 1, signed($fs[1]));
      pl_set_value($pe, 3, signed($fs[2])); push(@$pl, $pe);
    } elsif ($t == 141) { # push
      $pos += 1; push(@$pl, ['PUSH']); push(@$stk, {});
    } elsif ($t == 142) { # pop
      $pos += 1; push(@$pl, ['POP']); pop(@$stk);
      (@$stk) or return;
    } elsif (143 <= $t && $t <= 146) { # rightN
      ($pe, $pos) = vf_dvi_move1($dat, $pos, $stk, 'r', $t - 142);
      (defined $pe) or return; push(@$pl, $pe);
    } elsif ($t == 147) { # w0
      ($pe, $pos) = vf_dvi_move0($dat, $pos, $stk, 'w');
      (defined $pe) or return; push(@$pl, $pe);
    } elsif (148 <= $t && $t <= 151) { # wN
      ($pe, $pos) = vf_dvi_move1($dat, $pos, $stk, 'w', $t - 147);
      (defined $pe) or return; push(@$pl, $pe);
    } elsif ($t == 152) { # x0
      ($pe, $pos) = vf_dvi_move0($dat, $pos, $stk, 'x');
      (defined $pe) or return; push(@$pl, $pe);
    } elsif (153 <= $t && $t <= 156) { # xN
      ($pe, $pos) = vf_dvi_move1($dat, $pos, $stk, 'x', $t - 152);
      (defined $pe) or return; push(@$pl, $pe);
    } elsif (157 <= $t && $t <= 160) { # downN
      ($pe, $pos) = vf_dvi_move1($dat, $pos, $stk, 'd', $t - 156);
      (defined $pe) or return; push(@$pl, $pe);
    } elsif ($t == 161) { # y0
      ($pe, $pos) = vf_dvi_move0($dat, $pos, $stk, 'y');
      (defined $pe) or return; push(@$pl, $pe);
    } elsif (162 <= $t && $t <= 165) { # yN
      ($pe, $pos) = vf_dvi_move1($dat, $pos, $stk, 'y', $t - 161);
      (defined $pe) or return; push(@$pl, $pe);
    } elsif ($t == 166) { # z0
      ($pe, $pos) = vf_dvi_move0($dat, $pos, $stk, 'z');
      (defined $pe) or return; push(@$pl, $pe);
    } elsif (167 <= $t && $t <= 170) { # zN
      ($pe, $pos) = vf_dvi_move1($dat, $pos, $stk, 'z', $t - 166);
      (defined $pe) or return; push(@$pl, $pe);
    } elsif (171 <= $t && $t <= 234) { # fnt_numN
      $t -= 171; $pe = pl_cook(['SELECTFONT', 'D', 0]); $pos += 1;
      pl_set_value($pe, 1, $t); push(@$pl, $pe);
    } elsif (235 <= $t && $t <= 238) { # fntN
      $t -= 234; @fs = unpack("Ca$t", substr($dat, $pos));
      $pos += $t + 1; ($#fs == 1) or return;
      $pe = pl_cook(['SELECTFONT', 'D', 0]);
      pl_set_value($pe, 1, unpack_num($fs[1])); push(@$pl, $pe);
    } elsif (239 <= $t && $t <= 242) { # xxxN
      $t -= 238; @fs = unpack("Ca$t", substr($dat, $pos));
      $pos += $t + 1; ($#fs == 1) or return;
      $t = unpack_num($fs[1]); $u = substr($dat, $pos, $t);
      $pos += $t; ($t == length($u)) or return;
      $pe = vf_dvi_special($u); push(@$pl, $pe);
    } elsif ($t == 255) { # dir (JVF)
      @fs = unpack("CC", substr($dat, $pos));
      $pos += 2; ($#fs == 1) or return;
      $pe = pl_cook(['DIR', 'D', 0]);
      pl_set_value($pe, 1, $fs[1]); push(@$pl, $pe);
    } else { return; }
  }
  return $pl;
}

## vf_synerror($msg)
# Error messages in vf_parse.
sub vf_synerror
{
  return error("VF syntax error: $_[0]");
}

# for vf_dvi_move1 / vf_dvi_move0
my %vf_dvi_move = (
  r => 'MOVERIGHT', w => 'MOVERIGHT', x => 'MOVERIGHT',
  d => 'MOVEDOWN', y => 'MOVEDOWN', z => 'MOVEDOWN',
);

## vf_dvi_move1(...)
# Subcontractor of vf_dvi_parse.
sub vf_dvi_move1
{
  my ($dat, $pos, $stk, $r, $l) = @_; my ($t, $pe, @fs);
  @fs = unpack("Ca$l", substr($dat, $pos));
  $pos += $l + 1; ($#fs == 1) or return;
  $stk->[-1]{$r} = $t = unpack_snum($fs[1]);
  $pe = pl_cook([$vf_dvi_move{$r}, 'R', 0]);
  pl_set_value($pe, 1, $t);
  return ($pe, $pos);
}

## vf_dvi_move0(...)
# Subcontractor of vf_dvi_parse.
sub vf_dvi_move0
{
  my ($dat, $pos, $stk, $r) = @_; my ($t, $pe);
  (defined($t = $stk->[-1]{$r})) or return;
  $pe = pl_cook([$vf_dvi_move{$r}, 'R', 0]);
  pl_set_value($pe, 1, $t);
  return ($pe, $pos + 1);
}

## vf_dvi_special($dat)
# Subcontractor of vf_dvi_parse.
sub vf_dvi_special
{
  my ($dat) = @_; my ($t, $u, $pl); local ($errmsg);
  L1:{
    $t = "($dat)"; ($t !~ m/[^\x20-\x7e]/) or last;
    (defined($pl = pl_parse_list($t))) or last;
    (vf_issafe_list($pl)) or last;
    $u = pl_form_list($pl, -1); 
    return ['SPECIAL', $dat];
  }
  return ['SPEICALHEX', uc(unpack('H*', $dat))];
}

## vf_issafe_list($pl)
# Subcontractor of vf_dvi_special.
sub vf_issafe_list
{
  my ($pl) = @_; my ($ent);
  foreach $ent (@$pl) {
    if (ref $ent) {
      (vf_issafe_list($ent)) or return 0;
    } elsif ($ent =~ /^[CKDFOHR]$/ || $ent eq 'COMMENT') {
      return 0;
    }
  }
  return 1;
}

## vf_dvi_dumb_parse($dat)
# Subcontractor of dvi_parse.
sub vf_dvi_dumb_parse
{
  my ($dat) = @_;
 my ($t);
  $t = uc(unpack("H*", $dat));
  return ['MAP', ['DIRECTHEX', $t]];
}

##<*> vf_form($pl)
# Inverse of vf_parse.
sub vf_form
{
  my ($pl, $swfr) = @_;
  my ($t, $u, $v, $pe, @fs, @chds, @cfds, @ccps);
  @chds = (247, 202, "", 0, 10 * TU);
  foreach $pe (@$pl) {
    if ($pe->[0] eq 'VTITLE') {
      ($#$pe <= 1) or return vf_fsynerror("bad argument", $pe);
      (length($pe->[1]) <= 255)
        or return vf_fsynerror("VTITLE string too long", $pe->[1]);
      $chds[2] = $pe->[1];
    } elsif ($pe->[0] eq 'CHECKSUM') {
      ($#$pe == 2 && defined($t = pl_value($pe, 1)))
        or return vf_fsynerror("bad argument", $pe);
      $chds[3] = $t;
    } elsif ($pe->[0] eq 'DESIGNSIZE') {
      ($#$pe == 2 && defined($t = pl_value($pe, 1)))
        or return vf_fsynerror("bad argument", $pe);
      $chds[4] = $t;
    } elsif ($pe->[0] eq 'MAPFONT') {
     ($#$pe >= 2  && defined($t = pl_value($pe, 1)))
       or return vf_fsynerror("bad argument", $pe);
      ($u, $t) = pack_num($t);
      @fs = ($u + 242, $t, 0, TU, 10 * TU, 0, 0, '', '');
      foreach $pe (@{$pe}[3 .. $#$pe]) {
        if (!ref $pe) {
          return vf_fsynerror("unexpected bareword", $pe);
        } elsif ($pe->[0] eq 'FONTCHECKSUM') {
          ($#$pe == 2  && defined($fs[2] = pl_value($pe, 1)))
            or return vf_fsynerror("bad argument", $pe);
        } elsif ($pe->[0] eq 'FONTAT') {
          ($#$pe == 2  && defined($fs[3] = pl_value($pe, 1)))
            or return vf_fsynerror("bad argument", $pe);
        } elsif ($pe->[0] eq 'FONTDSIZE') {
          ($#$pe == 2  && defined($fs[4] = pl_value($pe, 1)))
            or return vf_fsynerror("bad argument", $pe);
        } elsif ($pe->[0] eq 'FONTAREA') {
          ($#$pe == 1  && length($pe->[1]) <= 255)
            or return vf_fsynerror("bad argument", $pe);
          $fs[7] = $pe->[1]; $fs[5] = length($pe->[1]);
        } elsif ($pe->[0] eq 'FONTNAME') {
          ($#$pe == 1  && length($pe->[1]) <= 255)
            or return vf_fsynerror("bad argument", $pe);
          $fs[8] = $pe->[1]; $fs[6] = length($pe->[1]);
        } elsif (!$swfr) {
          return vf_fsynerror("unknown property", $pe);
        }
      }
      push(@cfds, pack("Ca*NNNCCa*a*", @fs));
    } elsif ($pe->[0] eq 'CHARACTER') {
     ($#$pe >= 2  && defined($t = pl_value($pe, 1)))
       or return vf_fsynerror("bad argument", $pe);
      $v = 0; $u = undef;
      foreach $pe (@{$pe}[3 .. $#$pe]) {
        if (!ref $pe) {
          return vf_fsynerror("unexpected bareword", $pe);
        } elsif ($pe->[0] eq 'CHARWD') {
          ($#$pe == 2  && defined($v = pl_value($pe, 1)))
            or return vf_fsynerror("bad argument", $pe);
        } elsif ($pe->[0] eq 'MAP') {
          (defined($u = vf_dvi_form($pe, $t))) or return;
        } elsif (!$swfr) {
          return vf_fsynerror("unknown property", $pe);
        }
      }
      if (!defined $u) {
        $u = pl_cook(['MAP', ['SETCHAR']]);
        (defined($u = vf_dvi_form($u, $t))) or return;
      }
      if (0 <= $t && $t <= 255 && 0 <= $v && $v <= 0xFFFFFF &&
          length($u) <= 241) { # short form
        @fs = (length($u), $t, substr(pack('N', $v), 1), $u);
        push(@ccps, pack("CCa3a*", @fs));
      } else {
        @fs = (242, length($u), $t, $v, $u);
        push(@ccps, pack("CNNNa*", @fs));
      }
    } elsif (!$swfr) {
      return vf_fsynerror("unknown property", $pe);
    }
  }
  $t = pack("CCC/a*NN", @chds);
  $t = join('', $t, @cfds, @ccps);
  $t .= "\xf8" x (4 - length($t) % 4);
  return $t;
  
}

## vf_dvi_form($pl, $cc)
# Subcontractor of vf_form
sub vf_dvi_form
{
  my ($pl, $cc) = @_;
  my ($t, $u, $v, $l, $pe, $stk, @cnks);
  $stk = [{}];
  foreach $pe (@{$pl}[1 .. $#$pl]) {
    if ($pe->[0] eq 'SETCHAR') {
      if ($#$pe == 0) { $t = $cc; }
      elsif ($#$pe == 2 && defined($t = pl_value($pe, 1))) {
      } else { return vf_fsynerror("bad argument", $pe); }
      if (0 <= $t && $t <= 127) {
        push(@cnks, chr($t));
      } else {
        ($l, $t) = pack_num($t);
        push(@cnks, pack("Ca*", $l + 127, $t));
      }
    } elsif ($pe->[0] eq 'SETRULE') {
      ($#$pe == 4 && defined($t = pl_value($pe, 1)) &&
        defined($u = pl_value($pe, 3)))
        or return vf_fsynerror("bad argument", $pe);
      push(@cnks, pack("CNN", 132, $t, $u));
    } elsif ($pe->[0] eq 'PUSH') {
      push(@$stk, {}); push(@cnks, chr(141));
    } elsif ($pe->[0] eq 'POP') {
      pop(@$stk); (@$stk) or vf_fsynerror("cannot POP (char $cc)");
      push(@cnks, chr(142));
    } elsif ($pe->[0] eq 'MOVERIGHT') {
      (defined($t = vf_dvi_f_move($pe, $stk->[-1], 'r',+1))) or return;
      push(@cnks, $t);
    } elsif ($pe->[0] eq 'MOVELEFT') {
      (defined($t = vf_dvi_f_move($pe, $stk->[-1], 'r',-1))) or return;
      push(@cnks, $t);
    } elsif ($pe->[0] eq 'MOVEDOWN') {
      (defined($t = vf_dvi_f_move($pe, $stk->[-1], 'd',+1))) or return;
      push(@cnks, $t);
    } elsif ($pe->[0] eq 'MOVEUP') {
      (defined($t = vf_dvi_f_move($pe, $stk->[-1], 'd',-1))) or return;
      push(@cnks, $t);
    } elsif ($pe->[0] eq 'SELECTFONT') {
      ($#$pe == 2 && defined($t = pl_value($pe, 1)))
        or return vf_fsynerror("bad argument", $pe);
      if (0 <= $t && $t <= 63) {
        push(@cnks, chr($t + 171));
      } else {
        ($l, $t) = pack_num($t);
        push(@cnks, pack("Ca*", $l + 234, $t));
      }
    } elsif ($pe->[0] eq 'SPECIAL') {
      $t = pl_form_list($pe, -1);
      ($t =~ m|^\(SPECIAL\s?(.*)\)$|) or return fatal("vf_dvi_form");
      $u = $1; ($l, $t) = pack_num(length($u));
      push(@cnks, pack("Ca*a*", 238 + $l, $t, $u));
    } elsif ($pe->[0] eq 'SPECIALHEX') {
      ($u = join('', @{$pe}[1 .. $#$pl])) =~ s/\s+//g;
      ($u =~ m/^([0-9A-Fa-f]{2})+$/)
        or return vf_fsynerror("bad arguments", $pe);
      $u = pack("H*", $u); ($l, $t) = pack_num(length($u));
      push(@cnks, pack("Ca*a*", 238 + $l, $t, $u));
    } elsif ($pe->[0] eq 'DIR') {
      ($#$pe == 2 && defined($t = pl_value($pe, 1)) && $t <= 1)
        or return vf_fsynerror("bad argument", $pe);
      push(@cnks, pack("CC", 255, $t));
    } elsif ($pe->[0] eq 'DIRECTHEX') {
      ($u = join('', @{$pe}[1 .. $#$pl])) =~ s/\s+//g;
      ($u =~ m/^([0-9A-Fa-f]{2})+$/)
        or return vf_fsynerror("bad arguments", $pe);
      $u = pack("H*", $u); push(@cnks, $u);
    } else {
      return vf_fsynerror("unknown DVI operator: ", $pe);
    }
  }
  return join('', @cnks);
}

# for vf_simple_move_code()
our $vf_simple_move_code = 0;

## vf_simple_move_code($sw)
# Sets the value of $vf_simple_move_code. If it is true, then
# vf_form does not exploit w, x, y, z registers in compiling
# move operations in DVI code.
sub vf_simple_move_code { $vf_simple_move_code = $_[0]; }

## vf_dvi_f_move($pe, $stk, $r, $sgn)
sub vf_dvi_f_move
{
  my ($pe, $stk, $r, $sgn) = @_; my ($v, $l, $t, $w, $x, $b);
  ($#$pe == 2 && defined($v = pl_value($pe, 1)))
    or return vf_fsynerror("bad argument", $pe);
  if ($sgn < 0) { $v = -$v; }
  ($l, $t) = pack_snum($v);
  ($w, $x, $b) = ($r eq 'r') ? ('w', 'x', 142) : ('y', 'z', 156);
  if ($vf_simple_move_code) { $t = pack("Ca*", $b + $l, $t); }
  elsif (!exists $stk->{$w}) {
    $stk->{$w} = $v; $t = pack("Ca*", $b + $l + 5, $t);
  } elsif ($stk->{$w} == $v) { $t = chr($b + 5); }
  elsif (!exists $stk->{$x}) {
    $stk->{$x} = $v; $t = pack("Ca*", $b + $l + 10, $t);
  } elsif ($stk->{$x} == $v) { $t = chr($b + 10); }
  else { $t = pack("Ca*", $b + $l, $t); }
  return $t;
}

## vf_fsynerror($msg)
# Error messages in vf_form.
sub vf_fsynerror
{
  my ($msg, $pl) = @_;
  if (ref $pl) { $pl = pl_form_list($pl, -1); }
  return error("VPL syntax error: $msg: $pl");
}

##<*> vf_for_mapping($map, $fn, $rmt)
#
sub vf_for_mapping
{
  my ($map, $fn, $rmt) = @_; my ($e, $pe, $pe2, $pl, %hrmt);
  if (defined $rmt) {
    foreach (@$rmt) { $hrmt{$_->[0]} = $_; }
  }
  (defined($map = arraymap($map))) or return;
  $pl = pl_header_vpl({});
  push(@$pl, pl_cook(['MAPFONT', 'D', 0, ['FONTNAME', $fn]]));
  foreach $e (@$map) {
    if (defined $rmt && !defined $hrmt{$e->[0]}) { next; }
    $pe = pl_cook(['CHARACTER', 'C', 0,
                   ['MAP', ['SETCHAR', 'C', 0]]]);
    pl_set_value($pe, 1, $e->[0]);
    pl_set_value($pe->[3][1], 1, $e->[1]);
    if (defined $rmt) {
      $pe2 = pl_cook(['CHARWD', 'R', 0]);
      pl_set_real($pe2, 1, $hrmt{$e->[0]}[1]);
      splice(@$pe, 3, 0, $pe2);
    }
    push(@$pl, $pe);
  }
  return $pl;
}

##<*> vf_mapfont($pl, $fn)
#
sub vf_mapfont
{
  my ($pl, $fn) = @_; my ($t, $pe, $pe2, %hsh);
  foreach $pe (@$pl) {
    if ($pe->[0] eq 'MAPFONT') {
      ($pe2) = grep { $_->[0] eq 'FONTNAME' } (@{$pe}[3 .. $#$pe]);
      (defined $pe2) or next;
      if (!defined $fn) {
        $hsh{pl_value($pe, 1)} = $pe2->[1];
      } elsif (defined($t = pl_value($pe, 1)) && $t == $fn) {
        return $pe2->[1];
      }
    }
  }
  return (defined $fn) ? () : \%hsh;
}

##<*> vf_set_mapfont($pl, $fn, $fnam)
#
sub vf_set_mapfont
{
  my ($pl, $fn, $fnam) = @_; my ($t, $pe, $pe2, %hsh);
  foreach $pe (@$pl) {
    if ($pe->[0] eq 'MAPFONT' &&
        defined($t = pl_value($pe, 1)) && $t == $fn) {
      ($pe2) = grep { $_->[0] eq 'FONTNAME' } (@{$pe}[3 .. $#$pe]);
      (defined $pe2) or return 0;
      $pe2->[1] = $fnam; return 1;
    }
  }
  return 0;
}


######## 'jfm' section ########

##<*> jfm_use_uptex_tool($sw)
# Decides if upTeX tools are used to do jfm_parse. Here $sw means:
#   0: not use upTeX-pltotf
#   1: use upTeX-pltotf, invoked by 'pltotf'
#   2: use upTeX-pltotf, invoked by 'uppltotf'
our $jfm_use_uptex_tool = 0;
sub jfm_use_uptex_tool { $jfm_use_uptex_tool = $_[0]; }

##<*> jfm_parse($jfm)
# Converts JFM data $jfm to a pl-struct that describes a JPLZ.
sub jfm_parse
{
  my ($jfm) = @_; my ($cmd);
  if ($jfm_use_uptex_tool >= 1) {
    if ($jfm_use_uptex_tool == 2) { $cmd = 'uptftopl'; }
    return x_tftopl($jfm, $cmd);
  } else {
    return fatal("jfm_parse not yet implemented");
  }
}

##<*> jfm_form($pl)
# Converts a pl-struct $pl describing a JPLZ to JFM data.
sub jfm_form
{
  my ($pl) = @_; my ($cmd, $map, $jfm);
  if ($jfm_use_uptex_tool >= 1) {
    if ($jfm_use_uptex_tool == 2) { $cmd = 'uppltotf'; }
    (defined($pl = jfm_normalize($pl, 'U'))) or return;
    return x_pltotf($pl, $cmd);
  } else {
    (defined(($pl, $map) = jfm_preprocess($pl))) or return;
    $jfm = x_pltotf($pl, 'pltotf -kanji=jis');
    return jfm_postprocess($jfm, $map);
  }
}


## jfm_grab_cit($pl)
sub jfm_grab_cit
{
  my ($pl, $swnc) = @_; my ($t, $u, $k, $e, $cc, $ty, $pe);
  my (@pl2, %typ, @cit, @ccs);
  foreach $pe (@$pl) {
    if ($pe->[0] eq 'CHARSINTYPE') {
      pl_cook_list($pe);
      (defined($ty = pl_value($pe, 1))) or return;
      for ($k = 3; $k <= $#$pe; $k++) {
        $e = $pe->[$k];
        if (ref $e && $e->[0] eq CNUM) {
          (defined($cc = pl_value($pe, $k))) or return;
          $typ{$cc} = $ty; ++$k;
        } elsif ($e =~ m/^([JUX])([0-9A-Fa-f]{1,4})$/) {
          if (!$swnc && $1 ne 'X') {
            (defined($cc = jfm_injcode($1, hex($2)))) or return;
            $typ{$cc} = $ty;
          } else { $typ{hex($2)} = $ty; }
        } elsif ($e =~ m/^[^\x21-\x7e]/) {
          (defined($cc = jcode_ord($e)))
            or return error("malformed kanji character: ",
                            unpack('H*', $e));
            $typ{$cc} = $ty;
        } elsif (ref $e && $e->[0] eq 'CTRANGE') {
          (defined($t = pl_value($e, 1)) &&
           defined($u = pl_value($e, 3))) or return;
          foreach ($t .. $u) { $typ{$_} = $ty; }
        } else {
          return error("illegal element in CHARSINTYPE: ", $e);
        }
      }
      $cit[$ty] = $t = pl_cook(['CHARSINTYPE', 'D', 0]);
      pl_set_value($t, 1, $ty); push(@pl2, $t);
    } else { push(@pl2, $pe); }
  }
  return (\@pl2, \@cit, \%typ);
}

## jfm_normalize
sub jfm_normalize
{
  my ($pl, $f) = @_; my ($pl2, $cit, $typ, $cc, @ccs);
  if ($f !~ /^[JUX]$/) { $f = 'X'; }
  (defined(($pl2, $cit, $typ) = jfm_grab_cit($pl))) or return;
  @ccs = sort { $a <=> $b } (keys %$typ);
  foreach $cc (@ccs) {
    push(@{$cit->[$typ->{$cc}]}, sprintf("%s%04X", $f, $cc));
  }
  return $pl2;
}

## jfm_preprocess
# Subcontactor of jfm_form.
sub jfm_preprocess
{
  my ($pl) = @_; my ($pl2, $cit, $typ, $jc, $cc, @ccs, %map);
  (defined(($pl2, $cit, $typ) = jfm_grab_cit($pl))) or return;
  @ccs = sort { $a <=> $b } (keys %$typ);
  $jc = 0x2121;
  foreach $cc (@ccs) {
    push(@{$cit->[$typ->{$cc}]}, sprintf("J%04X", $jc));
    $map{$jc} = $cc; ++$jc;
    if (($jc & 0xFF) < 0x7F) { next; }
    if ($jc == 0x747F) {
      return error("too many chars have non-zero type");
    } elsif ($jc == 0x287F) { $jc = 0x3021; }
    else { $jc += 162; }
  }
  return ($pl2, \%map);
}
  # Valid codespace in pltotf: [21-28|30-74][21-7F] (7238 chars)

# for jfm_injcode
my %jfm_injc_sym = ( J => 'jis0208-raw', U => 'UTF-16BE' );

## jfm_injcode($f, $xc)
# Subcontactor of jfm_preprocess.
sub jfm_injcode
{
  my ($f, $xc) = @_; local ($jcode_ex);
  (defined($jcode_ex = $jfm_injc_sym{$f})) or return;
  if ($jcode_ex eq $jcode_in) { return $xc; }
  return jcode_ord(chr($xc >> 8) . chr($xc & 0xff));
}

## jfm_postprocess
# Subcontactor of jfm_form.
sub jfm_postprocess
{
  my ($jfm, $map) = @_; my ($k, $pct, $lct, $ct, @fs);
  @fs = unpack('nnnn', $jfm); 
  ($#fs == 3 && $fs[2] * 4 == length($jfm) && 
   ($fs[0] == 9 || $fs[0] == 11)) or return;
  $pct = $fs[3] * 4 + 28; $lct = $fs[1] * 4;
  $ct = substr($jfm, $pct, $lct); @fs = unpack('n*', $ct);
  for ($k = 2; $k <= $#fs; $k += 2) { $fs[$k] = $map->{$fs[$k]}; }
  $ct = pack('n*', @fs); 
  return substr($jfm, 0, $pct) . $ct . substr($jfm, $pct + $lct);
}

##<*> ttf_generate_jpl($ffnt, $id)
# Read the metric of a TTF file $ffnt (in case of TTC index is given
# by $id) and generates the pl-struct describing JPL for the metric.
# It returns ($pl, $dv) where $pl is the resulted pl-struct and $dv
# is the resulted rounding error of height/depth.
sub ttf_generate_jpl
{
  my ($ffnt, $id, $in) = @_; my ($t, $fam, $map, $out);
  my ($jpl, $vpl, $opl, $rmt, $jrmt, $imt, $vmt, $fd, $plcp);
  $out = { };
  if (defined $in->{family}) { $fam = $in->{family}; }
  else { ($fam = $ffnt) =~ s|.*/(.*)|$1|; };
  # JPL
  if (defined $in->{map}) {
    $map = $in->{map};
    (defined($jrmt = tfm_map_rmt($rmt, $map)))
      or return error("remapping failed");
  } else {
    $jrmt = $rmt; $map = [ ];
    foreach (@$rmt) { $map->{$_->[0]} = $_->[0]; }
  }
  ($out->{dw}, $out->{dv}, $imt, $vmt) =
   tfm_reduce_jpl_x($jrmt, $in->{balance});
  (defined $vmt) or return error("failed in length reduction");
  $t = { family => $fam, codingscheme => $in->{codingscheme} };
  (defined($fd = pl_fontdimen_jpl_rmt($rmt)) && 
   defined($jpl = pl_header($t, $fd)) &&
   defined($plcp = pl_char_part_jpl($imt, $vmt, 1)))
    or return error("JPL generation failed");
  push(@$jpl, @$plcp);
  # VF
  $jrmt = tfm_reduced_metric($imt, $vmt);
  $vpl = vf_for_mapping($map, $in->{ofm}, $jrmt);
  # OPL
  $t = { family => $fam, codingscheme => 'UNICODE' };
  (defined($fd = pl_fontdimen_opl_rmt($rmt)) &&
   defined($opl = pl_header_opl($t, $fd, 0)) &&
   defined($plcp = pl_char_part($rmt)))
    or return error("OPL generation failed");
  push(@$opl, @$plcp);
  # done
  return ($jpl, $vpl, $opl, $out);
}

##<*> ttf_generate_jpl($ffnt, $id, $in)
# Read the metric of a TTF file $ffnt (in case of TTC index is given
# by $id) and generates the composite PL for the metric. Here $in is
# a hash ref containing the following valid keys:
#   ofm: OFM name, recorded in the VF file
#   map: mapping table (either in array or hash form)
#   balance: balance factor applied in length reduction
#   family: font family name (only of informative use)
#   codingscheme: JFM codingscheme (only of informative use)
# It returns ($jpl, $vpl, $opl, $out), where the first three pl-structs
# describe JPL, VPL0 and OPL, and $out is a hash ref containing the
# following valid keys:
#   dw: maximum error of widths
#   dv: maximum error of heights/depths
sub ttf_generate_jpl_x
{
  my ($ffnt, $id, $in) = @_; my ($t, $fam, $map, $out);
  my ($jpl, $vpl, $opl, $rmt, $jrmt, $imt, $vmt, $fd, $plcp);
  $out = { };
  (defined($rmt = ttf_get_metric($ffnt, $id))) or return;
  if (defined $in->{family}) { $fam = $in->{family}; }
  else { ($fam = $ffnt) =~ s|.*/(.*)|$1|; };
  # JPL
  if (defined $in->{map}) {
    $map = $in->{map};
    (defined($jrmt = tfm_map_rmt($rmt, $map)))
      or return error("remapping failed");
  } else {
    $jrmt = $rmt; $map = { };
    foreach (@$rmt) { $map->{$_->[0]} = $_->[0]; }
  }
  ($out->{dw}, $out->{dv}, $imt, $vmt) =
   tfm_reduce_jpl_x($jrmt, $in->{balance});
  (defined $vmt) or return error("failed in length reduction");
  $t = { family => $fam, codingscheme => $in->{codingscheme} };
  (defined($fd = pl_fontdimen_jpl_rmt($rmt)) && 
   defined($jpl = pl_header($t, $fd)) &&
   defined($plcp = pl_char_part_jpl($imt, $vmt, 1)))
    or return error("JPL generation failed");
  push(@$jpl, @$plcp);
  # VF
  $jrmt = tfm_reduced_metric($imt, $vmt);
  $vpl = vf_for_mapping($map, $in->{ofm}, $jrmt);
  # OPL
  $t = { family => $fam, codingscheme => 'UNICODE' };
  (defined($fd = pl_fontdimen_opl_rmt($rmt)) &&
   defined($opl = pl_header_opl($t, $fd, 0)) &&
   defined($plcp = pl_char_part($rmt)))
    or return error("OPL generation failed");
  push(@$opl, @$plcp);
  # done
  return ($jpl, $vpl, $opl, $out);
}

######## all done
1;
# EOF
